# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import copy
import errno
import itertools
import logging
import math
import random
import uuid

import pickle  # nosec: B403
from copy import deepcopy
from contextlib import contextmanager

from array import array
from collections import defaultdict
from time import time

from swift.common import exceptions
from swift.common.ring.ring import RingData
from swift.common.ring.utils import tiers_for_dev, build_tier_tree, \
    validate_and_normalize_address, validate_replicas_by_tier, pretty_dev, \
    none_dev_id, calc_dev_id_bytes, BYTES_TO_TYPE_CODE, resize_array

# we can't store None's in the replica2part2dev array, so we high-jack
# the max value for magic to represent the part is not currently
# assigned to any device.
MAX_BALANCE = 999.99
MAX_BALANCE_GATHER_COUNT = 3


class RingValidationWarning(Warning):
    pass


@contextlib.contextmanager
def _set_random_seed(seed):
    # If random seed is set when entering this context then reset original
    # random state when exiting the context. This avoids a test calling this
    # method with a fixed seed value causing all subsequent tests to use a
    # repeatable random sequence.
    random_state = None
    if seed is not None:
        random_state = random.getstate()
        random.seed(seed)
    try:
        yield
    finally:
        if random_state:
            # resetting state rather than calling seed() eases unit testing
            random.setstate(random_state)


class RingBuilder(object):
    """
    Used to build swift.common.ring.RingData instances to be written to disk
    and used with swift.common.ring.Ring instances. See bin/swift-ring-builder
    for example usage.

    The instance variable devs_changed indicates if the device information has
    changed since the last balancing. This can be used by tools to know whether
    a rebalance request is an isolated request or due to added, changed, or
    removed devices.

    :param part_power: number of partitions = 2**part_power.
    :param replicas: number of replicas for each partition
    :param min_part_hours: minimum number of hours between partition changes
    """

    def __init__(self, part_power, replicas, min_part_hours):
        if part_power > 32:
            raise ValueError("part_power must be at most 32 (was %d)"
                             % (part_power,))
        if part_power < 0:
            raise ValueError("part_power must be at least 0 (was %d)"
                             % (part_power,))
        if replicas < 1:
            raise ValueError("replicas must be at least 1 (was %.6f)"
                             % (replicas,))
        if min_part_hours < 0:
            raise ValueError("min_part_hours must be non-negative (was %d)"
                             % (min_part_hours,))

        self.part_power = part_power
        self.next_part_power = None
        self.replicas = replicas
        self.min_part_hours = min_part_hours
        self.parts = 2 ** self.part_power
        self.devs = []
        self.devs_changed = False
        self.version = 0
        self.overload = 0.0
        self._id = None

        # _replica2part2dev maps from replica number to partition number to
        # device id. So, for a three replica, 2**23 ring, it's an array of
        # three 2**23 arrays of device ids (unsigned shorts). This can work a
        # bit faster than the 2**23 array of triplet arrays of device ids in
        # many circumstances. Making one big 2**23 * 3 array didn't seem to
        # have any speed change; though you're welcome to try it again (it was
        # a while ago, code-wise, when I last tried it).
        self._replica2part2dev = None

        # _last_part_moves is an array of unsigned bytes representing
        # the number of hours since a given partition was last moved.
        # This is used to guarantee we don't move a partition twice
        # within a given number of hours (24 is my usual test). Removing
        # a device overrides this behavior as it's assumed that's only
        # done because of device failure.
        self._last_part_moves = array('B', itertools.repeat(0, self.parts))
        # _part_moved_bitmap record parts have been moved
        self._part_moved_bitmap = None
        # _last_part_moves_epoch indicates the time the offsets in
        # _last_part_moves is based on.
        self._last_part_moves_epoch = 0

        self._last_part_gather_start = 0

        self._dispersion_graph = {}
        self.dispersion = 0.0
        self._remove_devs = []
        self._ring = None

        self.logger = logging.getLogger("swift.ring.builder")
        if not self.logger.handlers:
            self.logger.disabled = True
            # silence "no handler for X" error messages
            self.logger.addHandler(logging.NullHandler())

    @property
    def id(self):
        if self._id is None:
            # We don't automatically assign an id here because we want a caller
            # to explicitly know when a builder needs an id to be assigned. In
            # that case the caller must save the builder in order that a newly
            # assigned id is persisted.
            raise AttributeError(
                'id attribute has not been initialised by calling save()')
        return self._id

    @property
    def part_shift(self):
        return 32 - self.part_power

    @property
    def dev_id_bytes(self):
        if not self._replica2part2dev:
            max_dev_id = len(self.devs) - 1 if self.devs else 0
            return calc_dev_id_bytes(max_dev_id)
        return self._replica2part2dev[0].itemsize

    def set_dev_id_bytes(self, new_dev_id_bytes):
        if self._replica2part2dev:
            self._replica2part2dev = [
                resize_array(p2d, new_dev_id_bytes)
                for p2d in self._replica2part2dev]

    @property
    def dev_id_type_code(self):
        return BYTES_TO_TYPE_CODE[self.dev_id_bytes]

    @property
    def max_dev_id(self):
        return none_dev_id(self.dev_id_bytes) - 1

    @property
    def none_dev_id(self):
        return none_dev_id(self.dev_id_bytes)

    @property
    def ever_rebalanced(self):
        return self._replica2part2dev is not None

    def _set_part_moved(self, part):
        self._last_part_moves[part] = 0
        byte, bit = divmod(part, 8)
        self._part_moved_bitmap[byte] |= (128 >> bit)

    def _has_part_moved(self, part):
        byte, bit = divmod(part, 8)
        return bool(self._part_moved_bitmap[byte] & (128 >> bit))

    def _can_part_move(self, part):
        # if min_part_hours is zero then checking _last_part_moves will not
        # indicate if the part has already moved during the current rebalance,
        # but _has_part_moved will.
        return (self._last_part_moves[part] >= self.min_part_hours and
                not self._has_part_moved(part))

    @contextmanager
    def debug(self):
        """
        Temporarily enables debug logging, useful in tests, e.g.::

            with rb.debug():
                rb.rebalance()
        """
        old_val, self.logger.disabled = self.logger.disabled, False
        try:
            yield
        finally:
            self.logger.disabled = old_val

    @property
    def min_part_seconds_left(self):
        """Get the total seconds until a rebalance can be performed"""
        elapsed_seconds = int(time() - self._last_part_moves_epoch)
        return max((self.min_part_hours * 3600) - elapsed_seconds, 0)

    def weight_of_one_part(self):
        """
        Returns the weight of each partition as calculated from the
        total weight of all the devices.
        """
        try:
            return self.parts * self.replicas / \
                sum(d['weight'] for d in self._iter_devs())
        except ZeroDivisionError:
            raise exceptions.EmptyRingError('There are no devices in this '
                                            'ring, or all devices have been '
                                            'deleted')

    @classmethod
    def from_dict(cls, builder_data):
        b = cls(1, 1, 1)  # Dummy values
        b.copy_from(builder_data)
        return b

    def copy_from(self, builder):
        """
        Reinitializes this RingBuilder instance from data obtained from the
        builder dict given. Code example::

            b = RingBuilder(1, 1, 1)  # Dummy values
            b.copy_from(builder)

        This is to restore a RingBuilder that has had its b.to_dict()
        previously saved.
        """
        if hasattr(builder, 'devs'):
            self.part_power = builder.part_power
            self.next_part_power = builder.next_part_power
            self.replicas = builder.replicas
            self.min_part_hours = builder.min_part_hours
            self.parts = builder.parts
            self.devs = builder.devs
            self.devs_changed = builder.devs_changed
            self.overload = builder.overload
            self.version = builder.version
            self._replica2part2dev = builder._replica2part2dev
            self._last_part_moves_epoch = builder._last_part_moves_epoch
            if builder._last_part_moves is None:
                self._last_part_moves = array(
                    'B', itertools.repeat(0, self.parts))
            else:
                self._last_part_moves = builder._last_part_moves
            self._last_part_gather_start = builder._last_part_gather_start
            self._remove_devs = builder._remove_devs
            self._id = getattr(builder, '_id', None)
        else:
            self.part_power = builder['part_power']
            self.next_part_power = builder.get('next_part_power')
            self.replicas = builder['replicas']
            self.min_part_hours = builder['min_part_hours']
            self.parts = builder['parts']
            self.devs = builder['devs']
            self.devs_changed = builder['devs_changed']
            self.overload = builder.get('overload', 0.0)
            self.version = builder['version']
            self._replica2part2dev = builder['_replica2part2dev']
            self._last_part_moves_epoch = builder['_last_part_moves_epoch']
            if builder['_last_part_moves'] is None:
                self._last_part_moves = array(
                    'B', itertools.repeat(0, self.parts))
            else:
                self._last_part_moves = builder['_last_part_moves']
            self._last_part_gather_start = builder['_last_part_gather_start']
            self._dispersion_graph = builder.get('_dispersion_graph', {})
            self.dispersion = builder.get('dispersion')
            self._remove_devs = builder['_remove_devs']
            self._id = builder.get('id')
        self._ring = None

        # Old builders may not have a region defined for their devices, in
        # which case we default it to 1.
        for dev in self._iter_devs():
            dev.setdefault("region", 1)

        if not self._last_part_moves_epoch:
            self._last_part_moves_epoch = 0

    def __deepcopy__(self, memo):
        return type(self).from_dict(deepcopy(self.to_dict(), memo))

    def to_dict(self):
        """
        Returns a dict that can be used later with copy_from to
        restore a RingBuilder. swift-ring-builder uses this to
        pickle.dump the dict to a file and later load that dict into
        copy_from.
        """
        return {'part_power': self.part_power,
                'next_part_power': self.next_part_power,
                'replicas': self.replicas,
                'min_part_hours': self.min_part_hours,
                'parts': self.parts,
                'devs': self.devs,
                'devs_changed': self.devs_changed,
                'dev_id_bytes': self.dev_id_bytes,
                'version': self.version,
                'overload': self.overload,
                '_replica2part2dev': self._replica2part2dev,
                '_last_part_moves_epoch': self._last_part_moves_epoch,
                '_last_part_moves': self._last_part_moves,
                '_last_part_gather_start': self._last_part_gather_start,
                '_dispersion_graph': self._dispersion_graph,
                'dispersion': self.dispersion,
                '_remove_devs': self._remove_devs,
                'id': self._id}

    def change_min_part_hours(self, min_part_hours):
        """
        Changes the value used to decide if a given partition can be moved
        again. This restriction is to give the overall system enough time to
        settle a partition to its new location before moving it to yet another
        location. While no data would be lost if a partition is moved several
        times quickly, it could make that data unreachable for a short period
        of time.

        This should be set to at least the average full partition replication
        time. Starting it at 24 hours and then lowering it to what the
        replicator reports as the longest partition cycle is best.

        :param min_part_hours: new value for min_part_hours
        """
        self.min_part_hours = min_part_hours

    def set_replicas(self, new_replica_count):
        """
        Changes the number of replicas in this ring.

        If the new replica count is sufficiently different that
        self._replica2part2dev will change size, sets
        self.devs_changed. This is so tools like
        bin/swift-ring-builder can know to write out the new ring
        rather than bailing out due to lack of balance change.
        """
        old_slots_used = int(self.parts * self.replicas)
        new_slots_used = int(self.parts * new_replica_count)
        if old_slots_used != new_slots_used:
            self.devs_changed = True

        self.replicas = new_replica_count

    def set_overload(self, overload):
        self.overload = overload

    def get_ring(self):
        """
        Get the ring, or more specifically, the swift.common.ring.RingData.
        This ring data is the minimum required for use of the ring. The ring
        builder itself keeps additional data such as when partitions were last
        moved.
        """
        # We cache the self._ring value so multiple requests for it don't build
        # it multiple times. Be sure to set self._ring = None whenever the ring
        # will need to be rebuilt.
        if not self._ring:
            # Make devs list (with holes for deleted devices) and not including
            # builder-specific extra attributes.
            devs = [None] * len(self.devs)
            for dev in self._iter_devs():
                devs[dev['id']] = dict((k, v) for k, v in dev.items()
                                       if k not in ('parts', 'parts_wanted'))
            # Copy over the replica+partition->device assignments, the device
            # information, and the part_shift value (the number of bits to
            # shift an unsigned int >I right to obtain the partition for the
            # int).
            if not self._replica2part2dev:
                self._ring = RingData([], devs, self.part_shift,
                                      version=self.version)
            else:
                self._ring = \
                    RingData([array(self.dev_id_type_code, p2d)
                              for p2d in self._replica2part2dev],
                             devs, self.part_shift,
                             self.next_part_power,
                             self.version)
        return self._ring

    def add_dev(self, dev):
        """
        Add a device to the ring. This device dict should have a minimum of the
        following keys:

        ======  ===============================================================
        id      unique integer identifier amongst devices. Defaults to the next
                id if the 'id' key is not provided in the dict
        weight  a float of the relative weight of this device as compared to
                others; this indicates how many partitions the builder will try
                to assign to this device
        region  integer indicating which region the device is in
        zone    integer indicating which zone the device is in; a given
                partition will not be assigned to multiple devices within the
                same (region, zone) pair if there is any alternative
        ip      the ip address of the device
        port    the tcp port of the device
        device  the device's name on disk (sdb1, for example)
        meta    general use 'extra' field; for example: the online date, the
                hardware description
        ======  ===============================================================

        .. note::
            This will not rebalance the ring immediately as you may want to
            make multiple changes for a single rebalance.

        :param dev: device dict

        :returns: id of device (not used in the tree anymore, but unknown
                  users may depend on it)
        """
        if 'id' not in dev:
            dev['id'] = 0
            if self.devs:
                try:
                    dev['id'] = self.devs.index(None)
                except ValueError:
                    dev['id'] = len(self.devs)
        if dev['id'] < len(self.devs) and self.devs[dev['id']] is not None:
            raise exceptions.DuplicateDeviceError(
                'Duplicate device id: %d' % dev['id'])
        if dev['id'] > self.max_dev_id:
            self.set_dev_id_bytes(calc_dev_id_bytes(dev['id']))

        # Add holes to self.devs to ensure self.devs[dev['id']] will be the dev
        while dev['id'] >= len(self.devs):
            self.devs.append(None)
        required_keys = ('region', 'zone', 'ip', 'port', 'device', 'weight')
        missing = tuple(key for key in required_keys if key not in dev)
        if missing:
            raise ValueError('%r is missing required key(s): %s' % (
                dev, ', '.join(missing)))
        dev['weight'] = float(dev['weight'])
        dev['parts'] = 0
        dev.setdefault('meta', '')
        self.devs[dev['id']] = dev
        self.devs_changed = True
        self.version += 1
        return dev['id']

    def set_dev_weight(self, dev_id, weight):
        """
        Set the weight of a device. This should be called rather than just
        altering the weight key in the device dict directly, as the builder
        will need to rebuild some internal state to reflect the change.

        .. note::
            This will not rebalance the ring immediately as you may want to
            make multiple changes for a single rebalance.

        :param dev_id: device id
        :param weight: new weight for device
        """
        if any(dev_id == d['id'] for d in self._remove_devs):
            raise ValueError("Can not set weight of dev_id %s because it "
                             "is marked for removal" % (dev_id,))
        self.devs[dev_id]['weight'] = weight
        self.devs_changed = True
        self.version += 1

    def set_dev_region(self, dev_id, region):
        """
        Set the region of a device. This should be called rather than just
        altering the region key in the device dict directly, as the builder
        will need to rebuild some internal state to reflect the change.

        .. note::
            This will not rebalance the ring immediately as you may want to
            make multiple changes for a single rebalance.

        :param dev_id: device id
        :param region: new region for device
        """
        if any(dev_id == d['id'] for d in self._remove_devs):
            raise ValueError("Can not set region of dev_id %s because it "
                             "is marked for removal" % (dev_id,))
        self.devs[dev_id]['region'] = region
        self.devs_changed = True
        self.version += 1

    def set_dev_zone(self, dev_id, zone):
        """
        Set the zone of a device. This should be called rather than just
        altering the zone key in the device dict directly, as the builder
        will need to rebuild some internal state to reflect the change.

        .. note::
            This will not rebalance the ring immediately as you may want to
            make multiple changes for a single rebalance.

        :param dev_id: device id
        :param zone: new zone for device
        """
        if any(dev_id == d['id'] for d in self._remove_devs):
            raise ValueError("Can not set zone of dev_id %s because it "
                             "is marked for removal" % (dev_id,))
        self.devs[dev_id]['zone'] = zone
        self.devs_changed = True
        self.version += 1

    def remove_dev(self, dev_id):
        """
        Remove a device from the ring.

        .. note::
            This will not rebalance the ring immediately as you may want to
            make multiple changes for a single rebalance.

        :param dev_id: device id
        """
        dev = self.devs[dev_id]
        dev['weight'] = 0
        self._remove_devs.append(dev)
        self.devs_changed = True
        self.version += 1

    def rebalance(self, seed=None):
        """
        Rebalance the ring.

        This is the main work function of the builder, as it will assign and
        reassign partitions to devices in the ring based on weights, distinct
        zones, recent reassignments, etc.

        The process doesn't always perfectly assign partitions (that'd take a
        lot more analysis and therefore a lot more time -- I had code that did
        that before). Because of this, it keeps rebalancing until the device
        skew (number of partitions a device wants compared to what it has) gets
        below 1% or doesn't change by more than 1% (only happens with a ring
        that can't be balanced no matter what).

        :param seed: a value for the random seed (optional)
        :returns: (number_of_partitions_altered, resulting_balance,
                   number_of_removed_devices)
        """
        # count up the devs, and cache some stuff
        num_devices = 0
        for dev in self._iter_devs():
            dev['tiers'] = tiers_for_dev(dev)
            if dev['weight'] > 0:
                num_devices += 1
        if num_devices < self.replicas:
            raise exceptions.RingValidationError(
                "Replica count of %(replicas)s requires more "
                "than %(num_devices)s devices" % {
                    'replicas': self.replicas,
                    'num_devices': num_devices,
                })

        self._ring = None

        old_replica2part2dev = copy.deepcopy(self._replica2part2dev)

        if not self.ever_rebalanced:
            self.logger.debug("New builder; performing initial balance")

        self._update_last_part_moves()

        with _set_random_seed(seed):
            replica_plan = self._build_replica_plan()
            self._set_parts_wanted(replica_plan)

            assign_parts = defaultdict(list)
            # gather parts from replica count adjustment
            self._adjust_replica2part2dev_size(assign_parts)
            # gather parts from failed devices
            self._gather_parts_from_failed_devices(assign_parts)
            # gather parts for dispersion (N.B. this only picks up parts that
            # *must* disperse according to the replica plan)
            self._gather_parts_for_dispersion(assign_parts, replica_plan)
            removed_devs = self._remove_failed_devices()

            # we'll gather a few times, or until we archive the plan
            for gather_count in range(MAX_BALANCE_GATHER_COUNT):
                self._gather_parts_for_balance(assign_parts, replica_plan,
                                               # firsrt attempt go for disperse
                                               gather_count == 0)
                if not assign_parts:
                    # most likely min part hours
                    finish_status = 'Unable to finish'
                    break
                assign_parts_list = list(assign_parts.items())
                # shuffle the parts to be reassigned, we have no preference on
                # the order in which the replica plan is fulfilled.
                random.shuffle(assign_parts_list)
                # reset assign_parts map for next iteration
                assign_parts = defaultdict(list)

                num_part_replicas = sum(len(r) for p, r in assign_parts_list)
                self.logger.debug("Gathered %d parts", num_part_replicas)
                self._reassign_parts(assign_parts_list, replica_plan)
                self.logger.debug("Assigned %d parts", num_part_replicas)

                if not sum(d['parts_wanted'] < 0 for d in
                           self._iter_devs()):
                    finish_status = 'Finished'
                    break
            else:
                finish_status = 'Unable to finish'
            self.logger.debug(
                '%(status)s rebalance plan after %(count)s attempts',
                {'status': finish_status, 'count': gather_count + 1})

        self.devs_changed = False
        changed_parts = self._build_dispersion_graph(old_replica2part2dev)

        # clean up the cache
        for dev in self._iter_devs():
            dev.pop('tiers', None)

        return changed_parts, self.get_balance(), removed_devs

    def _build_dispersion_graph(self, old_replica2part2dev=None):
        """
        Build a dict of all tiers in the cluster to a list of the number of
        parts with a replica count at each index.  The values of the dict will
        be lists of length the maximum whole replica + 1 so that the
        graph[tier][3] is the number of parts within the tier with 3 replicas
        and graph [tier][0] is the number of parts not assigned in this tier.

        i.e.
        {
            <tier>: [
                <number_of_parts_with_0_replicas>,
                <number_of_parts_with_1_replicas>,
                ...
                <number_of_parts_with_n_replicas>,
                ],
            ...
        }

        :param old_replica2part2dev: if called from rebalance, the
            old_replica2part2dev can be used to count moved parts.

        :returns: number of parts with different assignments than
            old_replica2part2dev if provided
        """

        # Since we're going to loop over every replica of every part we'll
        # also count up changed_parts if old_replica2part2dev is passed in
        old_replica2part2dev = old_replica2part2dev or []
        # Compare the partition allocation before and after the rebalance
        # Only changed device ids are taken into account; devices might be
        # "touched" during the rebalance, but actually not really moved
        changed_parts = 0

        int_replicas = int(math.ceil(self.replicas))
        max_allowed_replicas = self._build_max_replicas_by_tier()
        parts_at_risk = 0

        dispersion_graph = {}
        # go over all the devices holding each replica part by part
        for part_id, dev_ids in enumerate(zip(*self._replica2part2dev)):
            # count the number of replicas of this part for each tier of each
            # device, some devices may have overlapping tiers!
            replicas_at_tier = defaultdict(int)
            for rep_id, dev in enumerate(iter(
                    self.devs[dev_id] for dev_id in dev_ids)):
                for tier in (dev.get('tiers') or tiers_for_dev(dev)):
                    replicas_at_tier[tier] += 1
                # IndexErrors will be raised if the replicas are increased or
                # decreased, and that actually means the partition has changed
                try:
                    old_device = old_replica2part2dev[rep_id][part_id]
                except IndexError:
                    changed_parts += 1
                    continue

                if old_device != dev['id']:
                    changed_parts += 1
            # update running totals for each tiers' number of parts with a
            # given replica count
            part_risk_depth = defaultdict(int)
            part_risk_depth[0] = 0
            for tier, replicas in replicas_at_tier.items():
                if tier not in dispersion_graph:
                    dispersion_graph[tier] = [self.parts] + [0] * int_replicas
                dispersion_graph[tier][0] -= 1
                dispersion_graph[tier][replicas] += 1
                if replicas > max_allowed_replicas[tier]:
                    part_risk_depth[len(tier)] += (
                        replicas - max_allowed_replicas[tier])
            # count each part-replica once at tier where dispersion is worst
            parts_at_risk += max(part_risk_depth.values())
        self._dispersion_graph = dispersion_graph
        self.dispersion = 100.0 * parts_at_risk / (self.parts * self.replicas)
        self.version += 1
        return changed_parts

    def validate(self, stats=False):
        """
        Validate the ring.

        This is a safety function to try to catch any bugs in the building
        process. It ensures partitions have been assigned to real devices,
        aren't doubly assigned, etc. It can also optionally check the even
        distribution of partitions across devices.

        :param stats: if True, check distribution of partitions across devices
        :returns: if stats is True, a tuple of (device_usage, worst_stat), else
                  (None, None). device_usage[dev_id] will equal the number of
                  partitions assigned to that device. worst_stat will equal the
                  number of partitions the worst device is skewed from the
                  number it should have.
        :raises RingValidationError: problem was found with the ring.
        """

        # "len" showed up in profiling, so it's just computed once.
        dev_len = len(self.devs)

        parts_on_devs = sum(d['parts'] for d in self._iter_devs())

        if not self._replica2part2dev:
            raise exceptions.RingValidationError(
                '_replica2part2dev empty; did you forget to rebalance?')

        parts_in_map = sum(len(p2d) for p2d in self._replica2part2dev)
        if parts_on_devs != parts_in_map:
            raise exceptions.RingValidationError(
                'All partitions are not double accounted for: %d != %d' %
                (parts_on_devs, parts_in_map))
        if stats:
            # dev_usage[dev_id] will equal the number of partitions assigned to
            # that device.
            dev_usage = array('I', (0 for _junk in range(dev_len)))
            for part2dev in self._replica2part2dev:
                for dev_id in part2dev:
                    dev_usage[dev_id] += 1

        for dev in self._iter_devs():
            if not isinstance(dev['port'], int):
                raise exceptions.RingValidationError(
                    "Device %d has port %r, which is not an integer." %
                    (dev['id'], dev['port']))

        int_replicas = int(math.ceil(self.replicas))
        rep2part_len = list(map(len, self._replica2part2dev))
        # check the assignments of each part's replicas
        for part in range(self.parts):
            devs_for_part = []
            for replica, part_len in enumerate(rep2part_len):
                if part_len <= part:
                    # last replica may be short on parts because of floating
                    # replica count
                    if replica + 1 < int_replicas:
                        raise exceptions.RingValidationError(
                            "The partition assignments of replica %r were "
                            "shorter than expected (%s < %s) - this should "
                            "only happen for the last replica" % (
                                replica,
                                len(self._replica2part2dev[replica]),
                                self.parts,
                            ))
                    break
                dev_id = self._replica2part2dev[replica][part]
                if dev_id == self.none_dev_id or dev_id >= dev_len or \
                        self.devs[dev_id] is None:
                    raise exceptions.RingValidationError(
                        "Partition %d, replica %d was not allocated "
                        "to a device." %
                        (part, replica))
                devs_for_part.append(dev_id)
            if len(devs_for_part) != len(set(devs_for_part)):
                raise exceptions.RingValidationError(
                    "The partition %s has been assigned to "
                    "duplicate devices %r" % (
                        part, devs_for_part))

        if stats:
            weight_of_one_part = self.weight_of_one_part()
            worst = 0
            for dev in self._iter_devs():
                if not dev['weight']:
                    if dev_usage[dev['id']]:
                        # If a device has no weight, but has partitions, then
                        # its overage is considered "infinity" and therefore
                        # always the worst possible. We show MAX_BALANCE for
                        # convenience.
                        worst = MAX_BALANCE
                        break
                    continue
                skew = abs(100.0 * dev_usage[dev['id']] /
                           (dev['weight'] * weight_of_one_part) - 100.0)
                if skew > worst:
                    worst = skew
            return dev_usage, worst
        return None, None

    def _build_balance_per_dev(self):
        """
        Build a map of <device_id> => <balance> where <balance> is a float
        representing the percentage difference from the desired amount of
        partitions a given device wants and the amount it has.

        N.B. this method only considers a device's weight and the parts
        assigned, not the parts wanted according to the replica plan.
        """
        weight_of_one_part = self.weight_of_one_part()
        balance_per_dev = {}
        for dev in self._iter_devs():
            if not dev['weight']:
                if dev['parts']:
                    # If a device has no weight, but has partitions, then its
                    # overage is considered "infinity" and therefore always the
                    # worst possible. We show MAX_BALANCE for convenience.
                    balance = MAX_BALANCE
                else:
                    balance = 0
            else:
                balance = 100.0 * dev['parts'] / (
                    dev['weight'] * weight_of_one_part) - 100.0
            balance_per_dev[dev['id']] = balance
        return balance_per_dev

    def get_balance(self):
        """
        Get the balance of the ring. The balance value is the highest
        percentage of the desired amount of partitions a given device
        wants. For instance, if the "worst" device wants (based on its
        weight relative to the sum of all the devices' weights) 123
        partitions and it has 124 partitions, the balance value would
        be 0.83 (1 extra / 123 wanted * 100 for percentage).

        :returns: balance of the ring
        """
        balance_per_dev = self._build_balance_per_dev()
        return max(abs(b) for b in balance_per_dev.values())

    def get_required_overload(self, weighted=None, wanted=None):
        """
        Returns the minimum overload value required to make the ring maximally
        dispersed.

        The required overload is the largest percentage change of any single
        device from its weighted replicanth to its wanted replicanth (note:
        under weighted devices have a negative percentage change) to archive
        dispersion - that is to say a single device that must be overloaded by
        5% is worse than 5 devices in a single tier overloaded by 1%.
        """
        weighted = weighted or self._build_weighted_replicas_by_tier()
        wanted = wanted or self._build_wanted_replicas_by_tier()
        max_overload = 0.0
        for dev in self._iter_devs():
            tier = (dev['region'], dev['zone'], dev['ip'], dev['id'])
            if not dev['weight']:
                if tier not in wanted or not wanted[tier]:
                    continue
                raise exceptions.RingValidationError(
                    'Device %s has zero weight and '
                    'should not want any replicas' % (tier,))
            required = (wanted[tier] - weighted[tier]) / weighted[tier]
            self.logger.debug('%(tier)s wants %(wanted)s and is weighted for '
                              '%(weight)s so therefore requires %(required)s '
                              'overload', {'tier': pretty_dev(dev),
                                           'wanted': wanted[tier],
                                           'weight': weighted[tier],
                                           'required': required})
            if required > max_overload:
                max_overload = required
        return max_overload

    def pretend_min_part_hours_passed(self):
        """
        Override min_part_hours by marking all partitions as having been moved
        255 hours ago and last move epoch to 'the beginning of time'. This can
        be used to force a full rebalance on the next call to rebalance.
        """
        self._last_part_moves_epoch = 0
        if not self._last_part_moves:
            return
        for part in range(self.parts):
            self._last_part_moves[part] = 0xff

    def get_part_devices(self, part):
        """
        Get the devices that are responsible for the partition,
        filtering out duplicates.

        :param part: partition to get devices for
        :returns: list of device dicts
        """
        devices = []
        for dev in self._devs_for_part(part):
            if dev not in devices:
                devices.append(dev)
        return devices

    def _iter_devs(self):
        """
        Returns an iterator all the non-None devices in the ring. Note that
        this means list(b._iter_devs())[some_id] may not equal b.devs[some_id];
        you will have to check the 'id' key of each device to obtain its
        dev_id.
        """
        for dev in self.devs:
            if dev is not None:
                yield dev

    def _build_tier2children(self):
        """
        Wrap helper build_tier_tree so exclude zero-weight devices.
        """
        return build_tier_tree(d for d in self._iter_devs() if d['weight'])

    def _set_parts_wanted(self, replica_plan):
        """
        Sets the parts_wanted key for each of the devices to the number of
        partitions the device wants based on its relative weight. This key is
        used to sort the devices according to "most wanted" during rebalancing
        to best distribute partitions. A negative parts_wanted indicates the
        device is "overweight" and wishes to give partitions away if possible.

        :param replica_plan: a dict of dicts, as returned from
                             _build_replica_plan, that maps
                             each tier to it's target replicanths.
        """
        tier2children = self._build_tier2children()

        parts_by_tier = defaultdict(int)

        def place_parts(tier, parts):
            parts_by_tier[tier] = parts
            sub_tiers = sorted(tier2children[tier])
            if not sub_tiers:
                return
            to_place = defaultdict(int)
            for t in sub_tiers:
                to_place[t] = min(parts, int(math.floor(
                    replica_plan[t]['target'] * self.parts)))
                parts -= to_place[t]

            # if there's some parts left over, just throw 'em about
            sub_tier_gen = itertools.cycle(sorted(
                sub_tiers, key=lambda t: replica_plan[t]['target']))
            while parts > 0:
                t = next(sub_tier_gen)
                to_place[t] += 1
                parts -= 1

            for t, p in to_place.items():
                place_parts(t, p)

        total_parts = int(self.replicas * self.parts)
        place_parts((), total_parts)

        # belts & suspenders/paranoia -  at every level, the sum of
        # parts_by_tier should be total_parts for the ring
        tiers = ['cluster', 'regions', 'zones', 'servers', 'devices']
        for i, tier_name in enumerate(tiers):
            parts_at_tier = sum(parts_by_tier[t] for t in parts_by_tier
                                if len(t) == i)
            if parts_at_tier != total_parts:
                raise exceptions.RingValidationError(
                    '%s != %s at tier %s' % (
                        parts_at_tier, total_parts, tier_name))

        for dev in self._iter_devs():
            if not dev['weight']:
                # With no weight, that means we wish to "drain" the device. So
                # we set the parts_wanted to a really large negative number to
                # indicate its strong desire to give up everything it has.
                dev['parts_wanted'] = -self.parts * self.replicas
            else:
                tier = (dev['region'], dev['zone'], dev['ip'], dev['id'])
                dev['parts_wanted'] = parts_by_tier[tier] - dev['parts']

    def _update_last_part_moves(self):
        """
        Updates how many hours ago each partition was moved based on the
        current time. The builder won't move a partition that has been moved
        more recently than min_part_hours.
        """
        self._part_moved_bitmap = bytearray(max(2 ** (self.part_power - 3), 1))
        elapsed_hours = int(time() - self._last_part_moves_epoch) // 3600
        if elapsed_hours <= 0:
            return
        for part in range(self.parts):
            # The "min(self._last_part_moves[part] + elapsed_hours, 0xff)"
            # which was here showed up in profiling, so it got inlined.
            last_plus_elapsed = self._last_part_moves[part] + elapsed_hours
            if last_plus_elapsed < 0xff:
                self._last_part_moves[part] = last_plus_elapsed
            else:
                self._last_part_moves[part] = 0xff
        self._last_part_moves_epoch = int(time())

    def _gather_parts_from_failed_devices(self, assign_parts):
        """
        Update the map of partition => [replicas] to be reassigned from
        removed devices.
        """
        # First we gather partitions from removed devices. Since removed
        # devices usually indicate device failures, we have no choice but to
        # reassign these partitions. However, we mark them as moved so later
        # choices will skip other replicas of the same partition if possible.

        gathered_parts = 0
        if self._remove_devs:
            dev_ids = [d['id'] for d in self._remove_devs if d['parts']]
            if dev_ids:
                for part, replica in self._each_part_replica():
                    dev_id = self._replica2part2dev[replica][part]
                    if dev_id in dev_ids:
                        self._replica2part2dev[replica][part] = \
                            self.none_dev_id
                        self._set_part_moved(part)
                        assign_parts[part].append(replica)
                        gathered_parts += 1
                        self.logger.debug(
                            "Gathered %d/%d from dev %d [dev removed]",
                            part, replica, dev_id)
        return gathered_parts

    def _remove_failed_devices(self):
        removed_devs = 0
        while self._remove_devs:
            remove_dev_id = self._remove_devs.pop()['id']
            self.logger.debug("Removing dev %d", remove_dev_id)
            self.devs[remove_dev_id] = None
            removed_devs += 1

        # Trim the dev list
        while self.devs and self.devs[-1] is None:
            self.devs.pop()

        if self.dev_id_bytes > 2:
            # Consider shrinking the device IDs themselves
            new_dev_id_bytes = self.dev_id_bytes // 2
            new_none_dev_id = none_dev_id(new_dev_id_bytes)
            # Only shrink if the IDs all fit in the lower half of the next size
            # down; this avoids excess churn when adding/removing devices near
            # the limit of a particular dev_id_bytes
            if len(self.devs) < new_none_dev_id // 2:
                self.set_dev_id_bytes(new_dev_id_bytes)

        return removed_devs

    def _adjust_replica2part2dev_size(self, to_assign):
        """
        Make sure that the lengths of the arrays in _replica2part2dev
        are correct for the current value of self.replicas.

        Example:
        self.part_power = 8
        self.replicas = 2.25

        self._replica2part2dev will contain 3 arrays: the first 2 of
        length 256 (2**8), and the last of length 64 (0.25 * 2**8).

        Update the mapping of partition => [replicas] that need assignment.
        """
        fractional_replicas, whole_replicas = math.modf(self.replicas)
        whole_replicas = int(whole_replicas)
        removed_parts = 0
        new_parts = 0

        desired_lengths = [self.parts] * whole_replicas
        if fractional_replicas:
            desired_lengths.append(int(self.parts * fractional_replicas))

        if self._replica2part2dev is not None:
            # If we crossed an integer threshold (say, 4.1 --> 4),
            # we'll have a partial extra replica clinging on here. Clean
            # up any such extra stuff.
            for part2dev in self._replica2part2dev[len(desired_lengths):]:
                for dev_id in part2dev:
                    dev_losing_part = self.devs[dev_id]
                    dev_losing_part['parts'] -= 1
                    removed_parts -= 1
            self._replica2part2dev = \
                self._replica2part2dev[:len(desired_lengths)]
        else:
            self._replica2part2dev = []

        for replica, desired_length in enumerate(desired_lengths):
            if replica < len(self._replica2part2dev):
                part2dev = self._replica2part2dev[replica]
                if len(part2dev) < desired_length:
                    # Not long enough: needs to be extended and the
                    # newly-added pieces assigned to devices.
                    for part in range(len(part2dev), desired_length):
                        to_assign[part].append(replica)
                        part2dev.append(self.none_dev_id)
                        new_parts += 1
                elif len(part2dev) > desired_length:
                    # Too long: truncate this mapping.
                    for part in range(desired_length, len(part2dev)):
                        dev_losing_part = self.devs[part2dev[part]]
                        dev_losing_part['parts'] -= 1
                        removed_parts -= 1
                    self._replica2part2dev[replica] = part2dev[:desired_length]
            else:
                # Mapping not present at all: make one up and assign
                # all of it.
                for part in range(desired_length):
                    to_assign[part].append(replica)
                    new_parts += 1
                self._replica2part2dev.append(
                    array(self.dev_id_type_code,
                          itertools.repeat(self.none_dev_id, desired_length)))

        self.logger.debug(
            "%d new parts and %d removed parts from replica-count change",
            new_parts, removed_parts)

    def _gather_parts_for_dispersion(self, assign_parts, replica_plan):
        """
        Update the map of partition => [replicas] to be reassigned from
        insufficiently-far-apart replicas.
        """
        # Now we gather partitions that are "at risk" because they aren't
        # currently sufficient spread out across the cluster.
        for part in range(self.parts):
            if (not self._can_part_move(part)):
                continue
            # First, add up the count of replicas at each tier for each
            # partition.
            replicas_at_tier = defaultdict(int)
            for dev in self._devs_for_part(part):
                for tier in dev['tiers']:
                    replicas_at_tier[tier] += 1

            # Now, look for partitions not yet spread out enough.
            undispersed_dev_replicas = []
            for replica in self._replicas_for_part(part):
                dev_id = self._replica2part2dev[replica][part]
                if dev_id == self.none_dev_id:
                    continue
                dev = self.devs[dev_id]
                if all(replicas_at_tier[tier] <=
                       replica_plan[tier]['max']
                       for tier in dev['tiers']):
                    continue
                undispersed_dev_replicas.append((dev, replica))

            if not undispersed_dev_replicas:
                continue

            undispersed_dev_replicas.sort(
                key=lambda dr: dr[0]['parts_wanted'])
            for dev, replica in undispersed_dev_replicas:
                # the min part hour check is ignored if and only if a device
                # has more than one replica of a part assigned to it - which
                # would have only been possible on rings built with an older
                # version of the code
                if (not self._can_part_move(part) and
                        not replicas_at_tier[dev['tiers'][-1]] > 1):
                    continue
                dev['parts_wanted'] += 1
                dev['parts'] -= 1
                assign_parts[part].append(replica)
                self.logger.debug(
                    "Gathered %d/%d from dev %s [dispersion]",
                    part, replica, pretty_dev(dev))
                self._replica2part2dev[replica][part] = self.none_dev_id
                for tier in dev['tiers']:
                    replicas_at_tier[tier] -= 1
                self._set_part_moved(part)

    def _gather_parts_for_balance_can_disperse(self, assign_parts, start,
                                               replica_plan):
        """
        Update the map of partition => [replicas] to be reassigned from
        overweight drives where the replicas can be better dispersed to
        another failure domain.

        :param assign_parts: the map of partition => [replica] to update
        :param start: offset into self.parts to begin search
        :param replica_plan: replicanth targets for tiers
        """
        tier2children = self._build_tier2children()
        parts_wanted_in_tier = defaultdict(int)
        for dev in self._iter_devs():
            wanted = max(dev['parts_wanted'], 0)
            for tier in dev['tiers']:
                parts_wanted_in_tier[tier] += wanted
        # Last, we gather partitions from devices that are "overweight" because
        # they have more partitions than their parts_wanted.
        for offset in range(self.parts):
            part = (start + offset) % self.parts
            if (not self._can_part_move(part)):
                continue
            # For each part we'll look at the devices holding those parts and
            # see if any are overweight, keeping track of replicas_at_tier as
            # we go
            overweight_dev_replica = []
            replicas_at_tier = defaultdict(int)
            for replica in self._replicas_for_part(part):
                dev_id = self._replica2part2dev[replica][part]
                if dev_id == self.none_dev_id:
                    continue
                dev = self.devs[dev_id]
                for tier in dev['tiers']:
                    replicas_at_tier[tier] += 1
                if dev['parts_wanted'] < 0:
                    overweight_dev_replica.append((dev, replica))

            if not overweight_dev_replica:
                continue

            overweight_dev_replica.sort(
                key=lambda dr: dr[0]['parts_wanted'])
            for dev, replica in overweight_dev_replica:
                if any(replica_plan[tier]['min'] <=
                       replicas_at_tier[tier] <
                       replica_plan[tier]['max']
                       for tier in dev['tiers']):
                    # we're stuck by replica plan
                    continue
                for t in reversed(dev['tiers']):
                    if replicas_at_tier[t] - 1 < replica_plan[t]['min']:
                        # we're stuck at tier t
                        break
                if sum(parts_wanted_in_tier[c]
                       for c in tier2children[t]
                       if c not in dev['tiers']) <= 0:
                    # we're stuck by weight
                    continue
                # this is the most overweight_device holding a replica
                # of this part that can shed it according to the plan
                dev['parts_wanted'] += 1
                dev['parts'] -= 1
                assign_parts[part].append(replica)
                self.logger.debug(
                    "Gathered %d/%d from dev %s [weight disperse]",
                    part, replica, pretty_dev(dev))
                self._replica2part2dev[replica][part] = self.none_dev_id
                for tier in dev['tiers']:
                    replicas_at_tier[tier] -= 1
                    parts_wanted_in_tier[tier] -= 1
                self._set_part_moved(part)
                break

    def _gather_parts_for_balance(self, assign_parts, replica_plan,
                                  disperse_first):
        """
        Gather parts that look like they should move for balance reasons.

        A simple gathers of parts that looks dispersible normally works out,
        we'll switch strategies if things don't seem to move.
        :param disperse_first: boolean, avoid replicas on overweight devices
                               that need to be there for dispersion
        """
        # pick a random starting point on the other side of the ring
        quarter_turn = (self.parts // 4)
        random_half = random.randint(0, self.parts // 2)
        start = (self._last_part_gather_start + quarter_turn +
                 random_half) % self.parts
        self.logger.debug('Gather start is %(start)s '
                          '(Last start was %(last_start)s)',
                          {'start': start,
                           'last_start': self._last_part_gather_start})
        self._last_part_gather_start = start

        if disperse_first:
            self._gather_parts_for_balance_can_disperse(
                assign_parts, start, replica_plan)
        self._gather_parts_for_balance_forced(assign_parts, start)

    def _gather_parts_for_balance_forced(self, assign_parts, start, **kwargs):
        """
        Update the map of partition => [replicas] to be reassigned from
        overweight drives without restriction, parts gathered from this method
        may be placed back onto devices that are no better (or worse) than the
        device from which they are gathered.

        This method allows devices to flop around enough to unlock replicas
        that would have otherwise potentially been locked because of
        dispersion - it should be used as a last resort.

        :param assign_parts: the map of partition => [replica] to update
        :param start: offset into self.parts to begin search
        """
        for offset in range(self.parts):
            part = (start + offset) % self.parts
            if (not self._can_part_move(part)):
                continue
            overweight_dev_replica = []
            for replica in self._replicas_for_part(part):
                dev_id = self._replica2part2dev[replica][part]
                if dev_id == self.none_dev_id:
                    continue
                dev = self.devs[dev_id]
                if dev['parts_wanted'] < 0:
                    overweight_dev_replica.append((dev, replica))

            if not overweight_dev_replica:
                continue

            overweight_dev_replica.sort(
                key=lambda dr: dr[0]['parts_wanted'])

            dev, replica = overweight_dev_replica[0]
            # this is the most overweight_device holding a replica of this
            # part we don't know where it's going to end up - but we'll
            # pick it up and hope for the best.
            dev['parts_wanted'] += 1
            dev['parts'] -= 1
            assign_parts[part].append(replica)
            self.logger.debug(
                "Gathered %d/%d from dev %s [weight forced]",
                part, replica, pretty_dev(dev))
            self._replica2part2dev[replica][part] = self.none_dev_id
            self._set_part_moved(part)

    def _reassign_parts(self, reassign_parts, replica_plan):
        """
        For an existing ring data set, partitions are reassigned similar to
        the initial assignment.

        The devices are ordered by how many partitions they still want and
        kept in that order throughout the process.

        The gathered partitions are iterated through, assigning them to
        devices according to the "most wanted" while keeping the replicas as
        "far apart" as possible.

        Two different regions are considered the farthest-apart things,
        followed by zones, then different ip within a zone; the
        least-far-apart things are different devices with the same ip in the
        same zone.

        :param reassign_parts: An iterable of (part, replicas_to_replace)
                               pairs. replicas_to_replace is an iterable of the
                               replica (an int) to replace for that partition.
                               replicas_to_replace may be shared for multiple
                               partitions, so be sure you do not modify it.
        """
        parts_available_in_tier = defaultdict(int)
        for dev in self._iter_devs():
            dev['sort_key'] = self._sort_key_for(dev)
            # Note: this represents how many partitions may be assigned to a
            # given tier (region/zone/server/disk). It does not take into
            # account how many partitions a given tier wants to shed.
            #
            # If we did not do this, we could have a zone where, at some
            # point during an assignment, number-of-parts-to-gain equals
            # number-of-parts-to-shed. At that point, no further placement
            # into that zone would occur since its parts_available_in_tier
            # would be 0. This would happen any time a zone had any device
            # with partitions to shed, which is any time a device is being
            # removed, which is a pretty frequent operation.
            wanted = max(dev['parts_wanted'], 0)
            for tier in dev['tiers']:
                parts_available_in_tier[tier] += wanted

        available_devs = \
            sorted((d for d in self._iter_devs() if d['weight']),
                   key=lambda x: x['sort_key'])

        tier2devs = defaultdict(list)
        tier2sort_key = defaultdict(tuple)
        tier2dev_sort_key = defaultdict(list)
        max_tier_depth = 0
        for dev in available_devs:
            for tier in dev['tiers']:
                tier2devs[tier].append(dev)  # <-- starts out sorted!
                tier2dev_sort_key[tier].append(dev['sort_key'])
                tier2sort_key[tier] = dev['sort_key']
                if len(tier) > max_tier_depth:
                    max_tier_depth = len(tier)

        tier2children_sets = build_tier_tree(available_devs)
        tier2children = defaultdict(list)
        tier2children_sort_key = {}
        tiers_list = [()]
        depth = 1
        while depth <= max_tier_depth:
            new_tiers_list = []
            for tier in tiers_list:
                child_tiers = list(tier2children_sets[tier])
                child_tiers.sort(key=tier2sort_key.__getitem__)
                tier2children[tier] = child_tiers
                tier2children_sort_key[tier] = map(
                    tier2sort_key.__getitem__, child_tiers)
                new_tiers_list.extend(child_tiers)
            tiers_list = new_tiers_list
            depth += 1

        for part, replace_replicas in reassign_parts:
            # always update part_moves for min_part_hours
            self._last_part_moves[part] = 0
            # count up where these replicas be
            replicas_at_tier = defaultdict(int)
            for dev in self._devs_for_part(part):
                for tier in dev['tiers']:
                    replicas_at_tier[tier] += 1

            for replica in replace_replicas:
                # Find a new home for this replica
                tier = ()
                # This used to be a cute, recursive function, but it's been
                # unrolled for performance.
                depth = 1
                while depth <= max_tier_depth:
                    # Choose the roomiest tier among those that don't
                    # already have their max replicas assigned according
                    # to the replica_plan.
                    candidates = [t for t in tier2children[tier] if
                                  replicas_at_tier[t] <
                                  replica_plan[t]['max']]

                    if not candidates:
                        raise Exception('no home for %s/%s %s' % (
                            part, replica, {t: (
                                replicas_at_tier[t],
                                replica_plan[t]['max'],
                            ) for t in tier2children[tier]}))
                    tier = max(candidates, key=lambda t:
                               parts_available_in_tier[t])

                    depth += 1

                dev = tier2devs[tier][-1]
                dev['parts_wanted'] -= 1
                dev['parts'] += 1
                for tier in dev['tiers']:
                    parts_available_in_tier[tier] -= 1
                    replicas_at_tier[tier] += 1

                self._replica2part2dev[replica][part] = dev['id']
                self.logger.debug(
                    "Placed %d/%d onto dev %s", part, replica, pretty_dev(dev))

        # Just to save memory and keep from accidental reuse.
        for dev in self._iter_devs():
            del dev['sort_key']

    @staticmethod
    def _sort_key_for(dev):
        return (dev['parts_wanted'], random.randint(0, 0xFFFF), dev['id'])

    def _build_max_replicas_by_tier(self, bound=math.ceil):
        """
        Returns a defaultdict of (tier: replica_count) for all tiers in the
        ring excluding zero weight devices.

        There will always be a () entry as the root of the structure, whose
        replica_count will equal the ring's replica_count.

        Then there will be (region,) entries for each region, indicating the
        maximum number of replicas the region might have for any given
        partition.

        Next there will be (region, zone) entries for each zone, indicating
        the maximum number of replicas in a given region and zone.  Anything
        greater than 1 indicates a partition at slightly elevated risk, as if
        that zone were to fail multiple replicas of that partition would be
        unreachable.

        Next there will be (region, zone, ip_port) entries for each node,
        indicating the maximum number of replicas stored on a node in a given
        region and zone.  Anything greater than 1 indicates a partition at
        elevated risk, as if that ip_port were to fail multiple replicas of
        that partition would be unreachable.

        Last there will be (region, zone, ip_port, device) entries for each
        device, indicating the maximum number of replicas the device shares
        with other devices on the same node for any given partition.
        Anything greater than 1 indicates a partition at serious risk, as the
        data on that partition will not be stored distinctly at the ring's
        replica_count.

        Example return dict for the common SAIO setup::

            {(): 3.0,
            (1,): 3.0,
            (1, 1): 1.0,
            (1, 1, '127.0.0.1:6210'): 1.0,
            (1, 1, '127.0.0.1:6210', 0): 1.0,
            (1, 2): 1.0,
            (1, 2, '127.0.0.1:6220'): 1.0,
            (1, 2, '127.0.0.1:6220', 1): 1.0,
            (1, 3): 1.0,
            (1, 3, '127.0.0.1:6230'): 1.0,
            (1, 3, '127.0.0.1:6230', 2): 1.0,
            (1, 4): 1.0,
            (1, 4, '127.0.0.1:6240'): 1.0,
            (1, 4, '127.0.0.1:6240', 3): 1.0}

        """
        # Used by walk_tree to know what entries to create for each recursive
        # call.
        tier2children = self._build_tier2children()

        def walk_tree(tier, replica_count):
            if len(tier) == 4:
                # special case for device, it's not recursive
                replica_count = min(1, replica_count)
            mr = {tier: replica_count}
            if tier in tier2children:
                subtiers = tier2children[tier]
                for subtier in subtiers:
                    submax = bound(float(replica_count) / len(subtiers))
                    mr.update(walk_tree(subtier, submax))
            return mr
        mr = defaultdict(float)
        mr.update(walk_tree((), self.replicas))
        return mr

    def _build_weighted_replicas_by_tier(self):
        """
        Returns a dict mapping <tier> => replicanths for all tiers in
        the ring based on their weights.
        """
        weight_of_one_part = self.weight_of_one_part()

        # assign each device some replicanths by weight (can't be > 1)
        weighted_replicas_for_dev = {}
        devices_with_room = []
        for dev in self._iter_devs():
            if not dev['weight']:
                continue
            weighted_replicas = (
                dev['weight'] * weight_of_one_part / self.parts)
            if weighted_replicas < 1:
                devices_with_room.append(dev['id'])
            else:
                weighted_replicas = 1
            weighted_replicas_for_dev[dev['id']] = weighted_replicas

        while True:
            remaining = self.replicas - sum(weighted_replicas_for_dev.values())
            if remaining < 1e-10:
                break
            devices_with_room = [d for d in devices_with_room if
                                 weighted_replicas_for_dev[d] < 1]
            rel_weight = remaining / sum(
                weighted_replicas_for_dev[d] for d in devices_with_room)
            for d in devices_with_room:
                weighted_replicas_for_dev[d] = min(
                    1, weighted_replicas_for_dev[d] * (rel_weight + 1))

        weighted_replicas_by_tier = defaultdict(float)
        for dev in self._iter_devs():
            if not dev['weight']:
                continue
            assigned_replicanths = weighted_replicas_for_dev[dev['id']]
            dev_tier = (dev['region'], dev['zone'], dev['ip'], dev['id'])
            for i in range(len(dev_tier) + 1):
                tier = dev_tier[:i]
                weighted_replicas_by_tier[tier] += assigned_replicanths

        # belts & suspenders/paranoia -  at every level, the sum of
        # weighted_replicas should be very close to the total number of
        # replicas for the ring
        validate_replicas_by_tier(self.replicas, weighted_replicas_by_tier)

        return weighted_replicas_by_tier

    def _build_wanted_replicas_by_tier(self):
        """
        Returns a defaultdict of (tier: replicanths) for all tiers in the ring
        based on unique-as-possible (full dispersion) with respect to their
        weights and device counts.

        N.B.  _build_max_replicas_by_tier calculates the upper bound on the
        replicanths each tier may hold irrespective of the weights of the
        tier; this method will calculate the minimum replicanth <=
        max_replicas[tier] that will still solve dispersion.  However, it is
        not guaranteed to return a fully dispersed solution if failure domains
        are over-weighted for their device count.
        """
        weighted_replicas = self._build_weighted_replicas_by_tier()
        dispersed_replicas = {
            t: {
                'min': math.floor(r),
                'max': math.ceil(r),
            } for (t, r) in
            self._build_max_replicas_by_tier(bound=float).items()
        }

        # watch out for device limited tiers
        num_devices = defaultdict(int)
        for d in self._iter_devs():
            if d['weight'] <= 0:
                continue
            for t in (d.get('tiers') or tiers_for_dev(d)):
                num_devices[t] += 1
            num_devices[()] += 1

        tier2children = self._build_tier2children()

        wanted_replicas = defaultdict(float)

        def place_replicas(tier, replicanths):
            if replicanths > num_devices[tier]:
                raise exceptions.RingValidationError(
                    'More replicanths (%s) than devices (%s) '
                    'in tier (%s)' % (replicanths, num_devices[tier], tier))
            wanted_replicas[tier] = replicanths
            sub_tiers = sorted(tier2children[tier])
            if not sub_tiers:
                return

            to_place = defaultdict(float)
            remaining = replicanths
            tiers_to_spread = sub_tiers
            device_limited = False

            while True:
                rel_weight = remaining / sum(weighted_replicas[t]
                                             for t in tiers_to_spread)
                for t in tiers_to_spread:
                    replicas = to_place[t] + (
                        weighted_replicas[t] * rel_weight)
                    if replicas < dispersed_replicas[t]['min']:
                        replicas = dispersed_replicas[t]['min']
                    elif (replicas > dispersed_replicas[t]['max'] and
                          not device_limited):
                        replicas = dispersed_replicas[t]['max']
                    if replicas > num_devices[t]:
                        replicas = num_devices[t]
                    to_place[t] = replicas

                remaining = replicanths - sum(to_place.values())

                if remaining < -1e-10:
                    tiers_to_spread = [
                        t for t in sub_tiers
                        if to_place[t] > dispersed_replicas[t]['min']
                    ]
                elif remaining > 1e-10:
                    tiers_to_spread = [
                        t for t in sub_tiers
                        if (num_devices[t] > to_place[t] <
                            dispersed_replicas[t]['max'])
                    ]
                    if not tiers_to_spread:
                        device_limited = True
                        tiers_to_spread = [
                            t for t in sub_tiers
                            if to_place[t] < num_devices[t]
                        ]
                else:
                    # remaining is "empty"
                    break

            for t in sub_tiers:
                self.logger.debug('Planning %s on %s',
                                  to_place[t], t)
                place_replicas(t, to_place[t])

        # place all replicas in the cluster tier
        place_replicas((), self.replicas)

        # belts & suspenders/paranoia -  at every level, the sum of
        # wanted_replicas should be very close to the total number of
        # replicas for the ring
        validate_replicas_by_tier(self.replicas, wanted_replicas)

        return wanted_replicas

    def _build_target_replicas_by_tier(self):
        """
        Build a map of <tier> => <target_replicas> accounting for device
        weights, unique-as-possible dispersion and overload.

        <tier> - a tuple, describing each tier in the ring topology
        <target_replicas> - a float, the target replicanths at the tier
        """
        weighted_replicas = self._build_weighted_replicas_by_tier()
        wanted_replicas = self._build_wanted_replicas_by_tier()
        max_overload = self.get_required_overload(weighted=weighted_replicas,
                                                  wanted=wanted_replicas)
        if max_overload <= 0.0:
            return wanted_replicas
        else:
            overload = min(self.overload, max_overload)
        self.logger.debug("Using effective overload of %f", overload)
        target_replicas = defaultdict(float)
        for tier, weighted in weighted_replicas.items():
            m = (wanted_replicas[tier] - weighted) / max_overload
            target_replicas[tier] = m * overload + weighted

        # belts & suspenders/paranoia -  at every level, the sum of
        # target_replicas should be very close to the total number
        # of replicas for the ring
        validate_replicas_by_tier(self.replicas, target_replicas)

        return target_replicas

    def _build_replica_plan(self):
        """
        Wraps return value of _build_target_replicas_by_tier to include
        pre-calculated min and max values for each tier.

        :returns: a dict, mapping <tier> => <replica_plan>, where
                  <replica_plan> is itself a dict

        <replica_plan> include at least the following keys:

            min - the minimum number of replicas at the tier
            target - the target replicanths at the tier
            max - the maximum number of replicas at the tier
        """
        # replica part-y planner!
        target_replicas = self._build_target_replicas_by_tier()
        replica_plan = defaultdict(
            lambda: {'min': 0, 'target': 0, 'max': 0})
        replica_plan.update({
            t: {
                'min': math.floor(r + 1e-10),
                'target': r,
                'max': math.ceil(r - 1e-10),
            } for (t, r) in
            target_replicas.items()
        })
        return replica_plan

    def _devs_for_part(self, part):
        """
        Returns a list of devices for a specified partition.

        Deliberately includes duplicates.
        """
        if self._replica2part2dev is None:
            return []
        devs = []
        for part2dev in self._replica2part2dev:
            if part >= len(part2dev):
                continue
            dev_id = part2dev[part]
            if dev_id == self.none_dev_id:
                continue
            devs.append(self.devs[dev_id])
        return devs

    def _replicas_for_part(self, part):
        """
        Returns a list of replicas for a specified partition.

        These can be used as indices into self._replica2part2dev
        without worrying about IndexErrors.
        """
        return [replica for replica, part2dev
                in enumerate(self._replica2part2dev)
                if part < len(part2dev)]

    def _each_part_replica(self):
        """
        Generator yielding every (partition, replica) pair in the ring.
        """
        for replica, part2dev in enumerate(self._replica2part2dev):
            for part in range(len(part2dev)):
                yield (part, replica)

    @classmethod
    def load(cls, builder_file, open=open, **kwargs):
        """
        Obtain RingBuilder instance of the provided builder file

        :param builder_file: path to builder file to load
        :return: RingBuilder instance
        """
        try:
            fp = open(builder_file, 'rb')
        except IOError as e:
            if e.errno == errno.ENOENT:
                raise exceptions.FileNotFoundError(
                    'Ring Builder file does not exist: %s' % builder_file)
            elif e.errno in [errno.EPERM, errno.EACCES]:
                raise exceptions.PermissionError(
                    'Ring Builder file cannot be accessed: %s' % builder_file)
            else:
                raise
        else:
            with fp:
                try:
                    builder = pickle.load(fp)  # nosec: B301
                except Exception:
                    # raise error during unpickling as UnPicklingError
                    raise exceptions.UnPicklingError(
                        'Ring Builder file is invalid: %s' % builder_file)

        if not hasattr(builder, 'devs'):
            builder_dict = builder
            builder = cls(1, 1, 1, **kwargs)
            builder.copy_from(builder_dict)

        if not hasattr(builder, '_id'):
            builder._id = None

        for dev in builder.devs:
            # really old rings didn't have meta keys
            if dev and 'meta' not in dev:
                dev['meta'] = ''
            # NOTE(akscram): An old ring builder file don't contain
            #                replication parameters.
            if dev:
                dev.setdefault('replication_ip', dev['ip'])
                dev.setdefault('replication_port', dev['port'])
        return builder

    def save(self, builder_file):
        """Serialize this RingBuilder instance to disk.

        :param builder_file: path to builder file to save
        """
        # We want to be sure the builder id's are persistent, so this is the
        # only place where the id is assigned. Newly created instances of this
        # class, or instances loaded from legacy builder files that have no
        # persisted id, must be saved in order for an id to be assigned.
        id_persisted = True
        if self._id is None:
            id_persisted = False
            self._id = uuid.uuid4().hex
        try:
            with open(builder_file, 'wb') as f:
                pickle.dump(self.to_dict(), f, protocol=2)
        except Exception:
            if not id_persisted:
                self._id = None
            raise

    def search_devs(self, search_values):
        """Search devices by parameters.

        :param search_values: a dictionary with search values to filter
                              devices, supported parameters are id,
                              region, zone, ip, port, replication_ip,
                              replication_port, device, weight, meta

        :returns: list of device dicts
        """
        matched_devs = []
        for dev in self.devs:
            if not dev:
                continue
            matched = True
            for key in ('id', 'region', 'zone', 'ip', 'port', 'replication_ip',
                        'replication_port', 'device', 'weight', 'meta'):
                if key in search_values:
                    value = search_values.get(key)
                    if value is not None:
                        if key == 'meta':
                            if value not in dev.get(key):
                                matched = False
                        elif key == 'ip' or key == 'replication_ip':
                            cdev = ''
                            try:
                                cdev = validate_and_normalize_address(
                                    dev.get(key, ''))
                            except ValueError:
                                pass
                            if cdev != value:
                                matched = False
                        elif dev.get(key) != value:
                            matched = False
            if matched:
                matched_devs.append(dev)
        return matched_devs

    def prepare_increase_partition_power(self):
        """
        Prepares a ring for partition power increase.

        This makes it possible to compute the future location of any object
        based on the next partition power.

        In this phase object servers should create hard links when finalizing a
        write to the new location as well. A relinker will be run after
        restarting object-servers, creating hard links to all existing objects
        in their future location.

        :returns: False if next_part_power was not set, otherwise True.
        """
        if self.next_part_power:
            return False
        self.next_part_power = self.part_power + 1
        self.version += 1
        return True

    def increase_partition_power(self):
        """
        Increases ring partition power by one.

        Devices will be assigned to partitions like this:

        OLD: 0, 3, 7, 5, 2, 1, ...
        NEW: 0, 0, 3, 3, 7, 7, 5, 5, 2, 2, 1, 1, ...

        :returns: False if next_part_power was not set or is equal to current
                  part_power, None if something went wrong, otherwise True.
        """

        if not self.next_part_power:
            return False

        if self.next_part_power != (self.part_power + 1):
            return False

        new_replica2part2dev = []
        for replica in self._replica2part2dev:
            new_replica = array(self.dev_id_type_code)
            for device in replica:
                new_replica.append(device)
                new_replica.append(device)  # append device a second time
            new_replica2part2dev.append(new_replica)
        self._replica2part2dev = new_replica2part2dev

        for device in self._iter_devs():
            device['parts'] *= 2

        # We need to update the time when a partition has been moved the last
        # time. Since this is an array of all partitions, we need to double it
        # too
        new_last_part_moves = []
        for partition in self._last_part_moves:
            new_last_part_moves.append(partition)
            new_last_part_moves.append(partition)
        self._last_part_moves = new_last_part_moves

        self.part_power = self.next_part_power
        self.parts *= 2
        self.version += 1
        return True

    def cancel_increase_partition_power(self):
        """
        Cancels a ring partition power increasement.

        This sets the next_part_power to the current part_power. Object
        replicators will still skip replication, and a cleanup is still
        required. Finally, a finish_increase_partition_power needs to be run.

        :returns: False if next_part_power was not set or is equal to current
                  part_power, otherwise True.
        """

        if not self.next_part_power:
            return False

        if self.next_part_power != (self.part_power + 1):
            return False

        self.next_part_power = self.part_power
        self.version += 1
        return True

    def finish_increase_partition_power(self):
        """Finish the partition power increase.

        The hard links from the old object locations should be removed by now.
        """
        if self.next_part_power and self.next_part_power == self.part_power:
            self.next_part_power = None
            self.version += 1
            return True
        return False



#!/usr/bin/env python
# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from optparse import OptionParser
from os.path import basename

from swift.common.ring import Ring
from swift.common.storage_policy import reload_storage_policies
from swift.common.utils import set_swift_dir
from swift.cli.info import (parse_get_node_args, print_item_locations,
                            InfoSystemExit)


def main():

    usage = '''
    Shows the nodes responsible for the item specified.
    Usage: %prog [-a] <ring.gz> <account> [<container> [<object>]]
    Or:    %prog [-a] <ring.gz> -p partition
    Or:    %prog [-a] -P policy_name <account> [<container> [<object>]]
    Or:    %prog [-a] -P policy_name -p partition
    Note: account, container, object can also be a single arg separated by /
    Example:
        $ %prog -a /etc/swift/account.ring.gz MyAccount
        Partition 5743883
        Hash 96ae332a60b58910784e4417a03e1ad0
        10.1.1.7:8000 sdd1
        10.1.9.2:8000 sdb1
        10.1.5.5:8000 sdf1
        10.1.5.9:8000 sdt1 # [Handoff]
    '''
    parser = OptionParser(usage)
    parser.add_option('-a', '--all', action='store_true',
                      help='Show all handoff nodes')
    parser.add_option('-p', '--partition', metavar='PARTITION',
                      help='Show nodes for a given partition')
    parser.add_option('-P', '--policy-name', dest='policy_name',
                      help='Specify which policy to use')
    parser.add_option('-d', '--swift-dir', default='/etc/swift',
                      dest='swift_dir', help='Path to swift directory')
    parser.add_option('-Q', '--quoted', action='store_true',
                      help='Assume swift paths are quoted')
    options, args = parser.parse_args()

    if set_swift_dir(options.swift_dir):
        reload_storage_policies()

    try:
        ring_path, args = parse_get_node_args(options, args)
    except InfoSystemExit as e:
        parser.print_help()
        sys.exit('ERROR: %s' % e)

    ring = ring_name = None
    if ring_path:
        ring_name = basename(ring_path)[:-len('.ring.gz')]
        ring = Ring(ring_path)

    try:
        print_item_locations(ring, ring_name, *args, **vars(options))
    except InfoSystemExit:
        sys.exit(1)


if __name__ == '__main__':
    main()



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from collections import defaultdict

from swift.common import exceptions
from swift.common import ring
from swift.common.ring.utils import (tiers_for_dev, build_tier_tree,
                                     validate_and_normalize_ip,
                                     validate_and_normalize_address,
                                     is_valid_hostname,
                                     is_local_device, parse_search_value,
                                     parse_search_values_from_opts,
                                     parse_change_values_from_opts,
                                     validate_args, parse_args,
                                     parse_builder_ring_filename_args,
                                     build_dev_from_opts, dispersion_report,
                                     parse_address, get_tier_name, pretty_dev,
                                     validate_replicas_by_tier)


class TestUtils(unittest.TestCase):

    def setUp(self):
        self.test_dev = {'region': 1, 'zone': 1, 'ip': '192.168.1.1',
                         'port': '6200', 'id': 0}

        def get_test_devs():
            dev0 = {'region': 1, 'zone': 1, 'ip': '192.168.1.1',
                    'port': '6200', 'id': 0}
            dev1 = {'region': 1, 'zone': 1, 'ip': '192.168.1.1',
                    'port': '6200', 'id': 1}
            dev2 = {'region': 1, 'zone': 1, 'ip': '192.168.1.1',
                    'port': '6200', 'id': 2}
            dev3 = {'region': 1, 'zone': 1, 'ip': '192.168.1.2',
                    'port': '6200', 'id': 3}
            dev4 = {'region': 1, 'zone': 1, 'ip': '192.168.1.2',
                    'port': '6200', 'id': 4}
            dev5 = {'region': 1, 'zone': 1, 'ip': '192.168.1.2',
                    'port': '6200', 'id': 5}
            dev6 = {'region': 1, 'zone': 2, 'ip': '192.168.2.1',
                    'port': '6200', 'id': 6}
            dev7 = {'region': 1, 'zone': 2, 'ip': '192.168.2.1',
                    'port': '6200', 'id': 7}
            dev8 = {'region': 1, 'zone': 2, 'ip': '192.168.2.1',
                    'port': '6200', 'id': 8}
            dev9 = {'region': 1, 'zone': 2, 'ip': '192.168.2.2',
                    'port': '6200', 'id': 9}
            dev10 = {'region': 1, 'zone': 2, 'ip': '192.168.2.2',
                     'port': '6200', 'id': 10}
            dev11 = {'region': 1, 'zone': 2, 'ip': '192.168.2.2',
                     'port': '6200', 'id': 11}
            return [dev0, dev1, dev2, dev3, dev4, dev5,
                    dev6, dev7, dev8, dev9, dev10, dev11]

        self.test_devs = get_test_devs()

    def test_tiers_for_dev(self):
        self.assertEqual(
            tiers_for_dev(self.test_dev),
            ((1,),
             (1, 1),
             (1, 1, '192.168.1.1'),
             (1, 1, '192.168.1.1', 0)))

    def test_build_tier_tree(self):
        ret = build_tier_tree(self.test_devs)
        self.assertEqual(len(ret), 8)
        self.assertEqual(ret[()], set([(1,)]))
        self.assertEqual(ret[(1,)], set([(1, 1), (1, 2)]))
        self.assertEqual(ret[(1, 1)],
                         set([(1, 1, '192.168.1.2'),
                              (1, 1, '192.168.1.1')]))
        self.assertEqual(ret[(1, 2)],
                         set([(1, 2, '192.168.2.2'),
                              (1, 2, '192.168.2.1')]))
        self.assertEqual(ret[(1, 1, '192.168.1.1')],
                         set([(1, 1, '192.168.1.1', 0),
                              (1, 1, '192.168.1.1', 1),
                              (1, 1, '192.168.1.1', 2)]))
        self.assertEqual(ret[(1, 1, '192.168.1.2')],
                         set([(1, 1, '192.168.1.2', 3),
                              (1, 1, '192.168.1.2', 4),
                              (1, 1, '192.168.1.2', 5)]))
        self.assertEqual(ret[(1, 2, '192.168.2.1')],
                         set([(1, 2, '192.168.2.1', 6),
                              (1, 2, '192.168.2.1', 7),
                              (1, 2, '192.168.2.1', 8)]))
        self.assertEqual(ret[(1, 2, '192.168.2.2')],
                         set([(1, 2, '192.168.2.2', 9),
                              (1, 2, '192.168.2.2', 10),
                              (1, 2, '192.168.2.2', 11)]))

    def test_is_valid_hostname(self):
        self.assertTrue(is_valid_hostname("local"))
        self.assertTrue(is_valid_hostname("test.test.com"))
        hostname = "test." * 51
        self.assertTrue(is_valid_hostname(hostname))
        hostname = hostname.rstrip('.')
        self.assertTrue(is_valid_hostname(hostname))
        hostname = hostname + "00"
        self.assertFalse(is_valid_hostname(hostname))
        self.assertFalse(is_valid_hostname("$blah#"))

    def test_is_local_device(self):
        # localhost shows up in whataremyips() output as "::1" for IPv6
        my_ips = ["127.0.0.1", "::1"]
        my_port = 6200
        self.assertTrue(is_local_device(my_ips, my_port,
                                        "127.0.0.1", my_port))
        self.assertTrue(is_local_device(my_ips, my_port,
                                        "::1", my_port))
        self.assertTrue(is_local_device(
            my_ips, my_port,
            "0000:0000:0000:0000:0000:0000:0000:0001", my_port))
        self.assertTrue(is_local_device(my_ips, my_port,
                                        "localhost", my_port))
        self.assertFalse(is_local_device(my_ips, my_port,
                                         "localhost", my_port + 1))
        self.assertFalse(is_local_device(my_ips, my_port,
                                         "127.0.0.2", my_port))
        # for those that don't have a local port
        self.assertTrue(is_local_device(my_ips, None,
                                        my_ips[0], None))

        # When servers_per_port is active, the "my_port" passed in is None
        # which means "don't include port in the determination of locality
        # because it's not reliable in this deployment scenario"
        self.assertTrue(is_local_device(my_ips, None,
                                        "127.0.0.1", 6666))
        self.assertTrue(is_local_device(my_ips, None,
                                        "::1", 6666))
        self.assertTrue(is_local_device(
            my_ips, None,
            "0000:0000:0000:0000:0000:0000:0000:0001", 6666))
        self.assertTrue(is_local_device(my_ips, None,
                                        "localhost", 6666))
        self.assertFalse(is_local_device(my_ips, None,
                                         "127.0.0.2", my_port))

    def test_validate_and_normalize_ip(self):
        ipv4 = "10.0.0.1"
        self.assertEqual(ipv4, validate_and_normalize_ip(ipv4))
        ipv6 = "fe80::204:61ff:fe9d:f156"
        self.assertEqual(ipv6, validate_and_normalize_ip(ipv6.upper()))
        hostname = "test.test.com"
        self.assertRaises(ValueError,
                          validate_and_normalize_ip, hostname)
        hostname = "$blah#"
        self.assertRaises(ValueError,
                          validate_and_normalize_ip, hostname)

    def test_validate_and_normalize_address(self):
        ipv4 = "10.0.0.1"
        self.assertEqual(ipv4, validate_and_normalize_address(ipv4))
        ipv6 = "fe80::204:61ff:fe9d:f156"
        self.assertEqual(ipv6, validate_and_normalize_address(ipv6.upper()))
        hostname = "test.test.com"
        self.assertEqual(hostname,
                         validate_and_normalize_address(hostname.upper()))
        hostname = "$blah#"
        self.assertRaises(ValueError,
                          validate_and_normalize_address, hostname)

    def test_validate_replicas_by_tier_close(self):
        one_ip_six_devices = \
            defaultdict(float,
                        {(): 4.0,
                         (0,): 4.0,
                         (0, 0): 4.0,
                         (0, 0, '127.0.0.1'): 4.0,
                         (0, 0, '127.0.0.1', 0): 0.6666666670,
                         (0, 0, '127.0.0.1', 1): 0.6666666668,
                         (0, 0, '127.0.0.1', 2): 0.6666666667,
                         (0, 0, '127.0.0.1', 3): 0.6666666666,
                         (0, 0, '127.0.0.1', 4): 0.6666666665,
                         (0, 0, '127.0.0.1', 5): 0.6666666664,
                         })
        try:
            validate_replicas_by_tier(4, one_ip_six_devices)
        except Exception as e:
            self.fail('one_ip_six_devices is invalid for %s' % e)

    def test_validate_replicas_by_tier_exact(self):
        three_regions_three_devices = \
            defaultdict(float,
                        {(): 3.0,
                         (0,): 1.0,
                         (0, 0): 1.0,
                         (0, 0, '127.0.0.1'): 1.0,
                         (0, 0, '127.0.0.1', 0): 1.0,
                         (1,): 1.0,
                         (1, 1): 1.0,
                         (1, 1, '127.0.0.1'): 1.0,
                         (1, 1, '127.0.0.1', 1): 1.0,
                         (2,): 1.0,
                         (2, 2): 1.0,
                         (2, 2, '127.0.0.1'): 1.0,
                         (2, 2, '127.0.0.1', 2): 1.0,
                         })
        try:
            validate_replicas_by_tier(3, three_regions_three_devices)
        except Exception as e:
            self.fail('three_regions_three_devices is invalid for %s' % e)

    def test_validate_replicas_by_tier_errors(self):
        pseudo_replicas = \
            defaultdict(float,
                        {(): 3.0,
                         (0,): 1.0,
                         (0, 0): 1.0,
                         (0, 0, '127.0.0.1'): 1.0,
                         (0, 0, '127.0.0.1', 0): 1.0,
                         (1,): 1.0,
                         (1, 1): 1.0,
                         (1, 1, '127.0.0.1'): 1.0,
                         (1, 1, '127.0.0.1', 1): 1.0,
                         (2,): 1.0,
                         (2, 2): 1.0,
                         (2, 2, '127.0.0.1'): 1.0,
                         (2, 2, '127.0.0.1', 2): 1.0,
                         })

        def do_test(bad_tier_key, bad_tier_name):
            # invalidate a copy of pseudo_replicas at given key and check for
            # an exception to be raised
            test_replicas = dict(pseudo_replicas)
            test_replicas[bad_tier_key] += 0.1  # <- this is not fair!
            with self.assertRaises(exceptions.RingValidationError) as ctx:
                validate_replicas_by_tier(3, test_replicas)
            self.assertEqual(
                '3.1 != 3 at tier %s' % bad_tier_name, str(ctx.exception))

        do_test((), 'cluster')
        do_test((1,), 'regions')
        do_test((0, 0), 'zones')
        do_test((2, 2, '127.0.0.1'), 'servers')
        do_test((1, 1, '127.0.0.1', 1), 'devices')

    def test_parse_search_value(self):
        res = parse_search_value('r0')
        self.assertEqual(res, {'region': 0})
        res = parse_search_value('r1')
        self.assertEqual(res, {'region': 1})
        res = parse_search_value('r1z2')
        self.assertEqual(res, {'region': 1, 'zone': 2})
        res = parse_search_value('d1')
        self.assertEqual(res, {'id': 1})
        res = parse_search_value('z1')
        self.assertEqual(res, {'zone': 1})
        res = parse_search_value('-127.0.0.1')
        self.assertEqual(res, {'ip': '127.0.0.1'})
        res = parse_search_value('127.0.0.1')
        self.assertEqual(res, {'ip': '127.0.0.1'})
        res = parse_search_value('-[127.0.0.1]:10001')
        self.assertEqual(res, {'ip': '127.0.0.1', 'port': 10001})
        res = parse_search_value(':10001')
        self.assertEqual(res, {'port': 10001})
        res = parse_search_value('R127.0.0.10')
        self.assertEqual(res, {'replication_ip': '127.0.0.10'})
        res = parse_search_value('R[127.0.0.10]:20000')
        self.assertEqual(res, {'replication_ip': '127.0.0.10',
                               'replication_port': 20000})
        res = parse_search_value('R:20000')
        self.assertEqual(res, {'replication_port': 20000})
        res = parse_search_value('/sdb1')
        self.assertEqual(res, {'device': 'sdb1'})
        res = parse_search_value('_meta1')
        self.assertEqual(res, {'meta': 'meta1'})
        self.assertRaises(ValueError, parse_search_value, 'OMGPONIES')

    def test_parse_search_values_from_opts(self):
        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "test.test.com",
             "--port", "6200",
             "--replication-ip", "r.test.com",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "change.test.test.com",
             "--change-port", "6201",
             "--change-replication-ip", "change.r.test.com",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        expected = {
            'id': 1,
            'region': 2,
            'zone': 3,
            'ip': "test.test.com",
            'port': 6200,
            'replication_ip': "r.test.com",
            'replication_port': 7000,
            'device': "sda3",
            'meta': "some meta data",
            'weight': 3.14159265359,
        }
        new_cmd_format, opts, args = validate_args(argv)
        search_values = parse_search_values_from_opts(opts)
        self.assertEqual(search_values, expected)

        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "127.0.0.1",
             "--port", "6200",
             "--replication-ip", "127.0.0.10",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "127.0.0.2",
             "--change-port", "6201",
             "--change-replication-ip", "127.0.0.20",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        expected = {
            'id': 1,
            'region': 2,
            'zone': 3,
            'ip': "127.0.0.1",
            'port': 6200,
            'replication_ip': "127.0.0.10",
            'replication_port': 7000,
            'device': "sda3",
            'meta': "some meta data",
            'weight': 3.14159265359,
        }
        new_cmd_format, opts, args = validate_args(argv)
        search_values = parse_search_values_from_opts(opts)
        self.assertEqual(search_values, expected)

        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "[127.0.0.1]",
             "--port", "6200",
             "--replication-ip", "[127.0.0.10]",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "[127.0.0.2]",
             "--change-port", "6201",
             "--change-replication-ip", "[127.0.0.20]",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        new_cmd_format, opts, args = validate_args(argv)
        search_values = parse_search_values_from_opts(opts)
        self.assertEqual(search_values, expected)

    def test_parse_change_values_from_opts(self):
        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "test.test.com",
             "--port", "6200",
             "--replication-ip", "r.test.com",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "change.test.test.com",
             "--change-port", "6201",
             "--change-replication-ip", "change.r.test.com",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        expected = {
            'ip': "change.test.test.com",
            'port': 6201,
            'replication_ip': "change.r.test.com",
            'replication_port': 7001,
            'device': "sdb3",
            'meta': "some meta data for change",
        }
        new_cmd_format, opts, args = validate_args(argv)
        search_values = parse_change_values_from_opts(opts)
        self.assertEqual(search_values, expected)

        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "127.0.0.1",
             "--port", "6200",
             "--replication-ip", "127.0.0.10",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "127.0.0.2",
             "--change-port", "6201",
             "--change-replication-ip", "127.0.0.20",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        expected = {
            'ip': "127.0.0.2",
            'port': 6201,
            'replication_ip': "127.0.0.20",
            'replication_port': 7001,
            'device': "sdb3",
            'meta': "some meta data for change",
        }
        new_cmd_format, opts, args = validate_args(argv)
        search_values = parse_change_values_from_opts(opts)
        self.assertEqual(search_values, expected)

        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "[127.0.0.1]",
             "--port", "6200",
             "--replication-ip", "[127.0.0.10]",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "[127.0.0.2]",
             "--change-port", "6201",
             "--change-replication-ip", "[127.0.0.20]",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        new_cmd_format, opts, args = validate_args(argv)
        search_values = parse_change_values_from_opts(opts)
        self.assertEqual(search_values, expected)

    def test_validate_args(self):
        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "test.test.com",
             "--port", "6200",
             "--replication-ip", "r.test.com",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "change.test.test.com",
             "--change-port", "6201",
             "--change-replication-ip", "change.r.test.com",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertTrue(new_cmd_format)
        self.assertEqual(opts.id, 1)
        self.assertEqual(opts.region, 2)
        self.assertEqual(opts.zone, 3)
        self.assertEqual(opts.ip, "test.test.com")
        self.assertEqual(opts.port, 6200)
        self.assertEqual(opts.replication_ip, "r.test.com")
        self.assertEqual(opts.replication_port, 7000)
        self.assertEqual(opts.device, "sda3")
        self.assertEqual(opts.meta, "some meta data")
        self.assertEqual(opts.weight, 3.14159265359)
        self.assertEqual(opts.change_ip, "change.test.test.com")
        self.assertEqual(opts.change_port, 6201)
        self.assertEqual(opts.change_replication_ip, "change.r.test.com")
        self.assertEqual(opts.change_replication_port, 7001)
        self.assertEqual(opts.change_device, "sdb3")
        self.assertEqual(opts.change_meta, "some meta data for change")

    def test_validate_args_new_cmd_format(self):
        argv = \
            ["--id", "0", "--region", "0", "--zone", "0",
             "--ip", "",
             "--port", "0",
             "--replication-ip", "",
             "--replication-port", "0",
             "--device", "",
             "--meta", "",
             "--weight", "0",
             "--change-ip", "",
             "--change-port", "0",
             "--change-replication-ip", "",
             "--change-replication-port", "0",
             "--change-device", "",
             "--change-meta", ""]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertTrue(new_cmd_format)

        argv = \
            ["--id", None, "--region", None, "--zone", None,
             "--ip", "",
             "--port", "0",
             "--replication-ip", "",
             "--replication-port", "0",
             "--device", "",
             "--meta", "",
             "--weight", None,
             "--change-ip", "change.test.test.com",
             "--change-port", "6201",
             "--change-replication-ip", "change.r.test.com",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertFalse(new_cmd_format)

        argv = \
            ["--id", "0"]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertTrue(new_cmd_format)
        argv = \
            ["--region", "0"]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertTrue(new_cmd_format)
        argv = \
            ["--zone", "0"]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertTrue(new_cmd_format)
        argv = \
            ["--weight", "0"]
        new_cmd_format, opts, args = validate_args(argv)
        self.assertTrue(new_cmd_format)

    def test_parse_args(self):
        argv = \
            ["--id", "1", "--region", "2", "--zone", "3",
             "--ip", "test.test.com",
             "--port", "6200",
             "--replication-ip", "r.test.com",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359",
             "--change-ip", "change.test.test.com",
             "--change-port", "6201",
             "--change-replication-ip", "change.r.test.com",
             "--change-replication-port", "7001",
             "--change-device", "sdb3",
             "--change-meta", "some meta data for change"]

        opts, args = parse_args(argv)
        self.assertEqual(opts.id, 1)
        self.assertEqual(opts.region, 2)
        self.assertEqual(opts.zone, 3)
        self.assertEqual(opts.ip, "test.test.com")
        self.assertEqual(opts.port, 6200)
        self.assertEqual(opts.replication_ip, "r.test.com")
        self.assertEqual(opts.replication_port, 7000)
        self.assertEqual(opts.device, "sda3")
        self.assertEqual(opts.meta, "some meta data")
        self.assertEqual(opts.weight, 3.14159265359)
        self.assertEqual(opts.change_ip, "change.test.test.com")
        self.assertEqual(opts.change_port, 6201)
        self.assertEqual(opts.change_replication_ip, "change.r.test.com")
        self.assertEqual(opts.change_replication_port, 7001)
        self.assertEqual(opts.change_device, "sdb3")
        self.assertEqual(opts.change_meta, "some meta data for change")
        self.assertEqual(len(args), 0)

    def test_parse_builder_ring_filename_args(self):
        args = 'swift-ring-builder object.builder write_ring'
        self.assertEqual((
            'object.builder', 'object.ring.gz'
        ), parse_builder_ring_filename_args(args.split()))
        args = 'swift-ring-builder container.ring.gz write_builder'
        self.assertEqual((
            'container.builder', 'container.ring.gz'
        ), parse_builder_ring_filename_args(args.split()))
        # builder name arg should always fall through
        args = 'swift-ring-builder test create'
        self.assertEqual((
            'test', 'test.ring.gz'
        ), parse_builder_ring_filename_args(args.split()))
        args = 'swift-ring-builder my.file.name create'
        self.assertEqual((
            'my.file.name', 'my.file.name.ring.gz'
        ), parse_builder_ring_filename_args(args.split()))

    def test_build_dev_from_opts(self):
        argv = \
            ["--region", "0", "--zone", "3",
             "--ip", "test.test.com",
             "--port", "6200",
             "--replication-ip", "r.test.com",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359"]
        expected = {
            'region': 0,
            'zone': 3,
            'ip': "test.test.com",
            'port': 6200,
            'replication_ip': "r.test.com",
            'replication_port': 7000,
            'device': "sda3",
            'meta': "some meta data",
            'weight': 3.14159265359,
        }
        opts, args = parse_args(argv)
        device = build_dev_from_opts(opts)
        self.assertEqual(device, expected)

        argv = \
            ["--region", "2", "--zone", "3",
             "--ip", "[test.test.com]",
             "--port", "6200",
             "--replication-ip", "[r.test.com]",
             "--replication-port", "7000",
             "--device", "sda3",
             "--meta", "some meta data",
             "--weight", "3.14159265359"]
        opts, args = parse_args(argv)
        self.assertRaises(ValueError, build_dev_from_opts, opts)

        argv = \
            ["--region", "2", "--zone", "3",
             "--ip", "[test.test.com]",
             "--port", "6200",
             "--replication-ip", "[r.test.com]",
             "--replication-port", "7000",
             "--meta", "some meta data",
             "--weight", "3.14159265359"]
        opts, args = parse_args(argv)
        self.assertRaises(ValueError, build_dev_from_opts, opts)

    def test_replication_defaults(self):
        args = '-r 1 -z 1 -i 127.0.0.1 -p 6010 -d d1 -w 100'.split()
        opts, _ = parse_args(args)
        device = build_dev_from_opts(opts)
        expected = {
            'device': 'd1',
            'ip': '127.0.0.1',
            'meta': '',
            'port': 6010,
            'region': 1,
            'replication_ip': '127.0.0.1',
            'replication_port': 6010,
            'weight': 100.0,
            'zone': 1,
        }
        self.assertEqual(device, expected)

        args = '-r 1 -z 1 -i test.com -p 6010 -d d1 -w 100'.split()
        opts, _ = parse_args(args)
        device = build_dev_from_opts(opts)
        expected = {
            'device': 'd1',
            'ip': 'test.com',
            'meta': '',
            'port': 6010,
            'region': 1,
            'replication_ip': 'test.com',
            'replication_port': 6010,
            'weight': 100.0,
            'zone': 1,
        }
        self.assertEqual(device, expected)

    def test_dispersion_report(self):
        rb = ring.RingBuilder(8, 3, 0)
        rb.add_dev({'id': 0, 'region': 1, 'zone': 0, 'weight': 100,
                    'ip': '127.0.0.0', 'port': 10000, 'device': 'sda1'})
        rb.add_dev({'id': 3, 'region': 1, 'zone': 0, 'weight': 100,
                    'ip': '127.0.0.0', 'port': 10000, 'device': 'sdb1'})
        rb.add_dev({'id': 4, 'region': 1, 'zone': 0, 'weight': 100,
                    'ip': '127.0.0.0', 'port': 10000, 'device': 'sdc1'})
        rb.add_dev({'id': 5, 'region': 1, 'zone': 0, 'weight': 100,
                    'ip': '127.0.0.0', 'port': 10000, 'device': 'sdd1'})

        rb.add_dev({'id': 1, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.1', 'port': 10001, 'device': 'sda1'})
        rb.add_dev({'id': 6, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.1', 'port': 10001, 'device': 'sdb1'})
        rb.add_dev({'id': 7, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.1', 'port': 10001, 'device': 'sdc1'})
        rb.add_dev({'id': 8, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.1', 'port': 10001, 'device': 'sdd1'})

        rb.add_dev({'id': 2, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.2', 'port': 10002, 'device': 'sda1'})
        rb.add_dev({'id': 9, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.2', 'port': 10002, 'device': 'sdb1'})
        rb.add_dev({'id': 10, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.2', 'port': 10002, 'device': 'sdc1'})
        rb.add_dev({'id': 11, 'region': 1, 'zone': 1, 'weight': 200,
                    'ip': '127.0.0.2', 'port': 10002, 'device': 'sdd1'})

        # this ring is pretty volatile and the assertions are pretty brittle
        # so we use a specific seed
        rb.rebalance(seed=100)
        rb.validate()

        self.assertEqual(rb.dispersion, 16.796875)
        report = dispersion_report(rb)
        self.assertEqual(report['worst_tier'], 'r1z1-127.0.0.1')
        self.assertEqual(report['max_dispersion'], 20.967741935483872)

        def build_tier_report(max_replicas, placed_parts, dispersion,
                              replicas):
            return {
                'max_replicas': max_replicas,
                'placed_parts': placed_parts,
                'dispersion': dispersion,
                'replicas': replicas,
            }

        # every partition has at least two replicas in this zone, unfortunately
        # sometimes they're both on the same server.
        expected = [
            ['r1z1', build_tier_report(
                2, 621, 17.55233494363929, [0, 0, 147, 109])],
            ['r1z1-127.0.0.1', build_tier_report(
                1, 310, 20.967741935483872, [11, 180, 65, 0])],
            ['r1z1-127.0.0.2', build_tier_report(
                1, 311, 20.578778135048232, [9, 183, 64, 0])],
        ]
        report = dispersion_report(rb, 'r1z1[^/]*$', verbose=True)
        graph = report['graph']
        for i, (expected_key, expected_report) in enumerate(expected):
            key, report = graph[i]
            self.assertEqual(
                (key, report),
                (expected_key, expected_report)
            )

        # overcompensate in r1z0
        rb.add_dev({'id': 12, 'region': 1, 'zone': 0, 'weight': 500,
                    'ip': '127.0.0.3', 'port': 10003, 'device': 'sda1'})
        rb.add_dev({'id': 13, 'region': 1, 'zone': 0, 'weight': 500,
                    'ip': '127.0.0.3', 'port': 10003, 'device': 'sdb1'})
        rb.add_dev({'id': 14, 'region': 1, 'zone': 0, 'weight': 500,
                    'ip': '127.0.0.3', 'port': 10003, 'device': 'sdc1'})
        rb.add_dev({'id': 15, 'region': 1, 'zone': 0, 'weight': 500,
                    'ip': '127.0.0.3', 'port': 10003, 'device': 'sdd1'})

        # when the biggest tier has the smallest devices things get ugly
        # can't move all the part-replicas in one rebalance
        rb.rebalance(seed=100)
        report = dispersion_report(rb, verbose=True)
        self.assertEqual(rb.dispersion, 2.8645833333333335)
        self.assertEqual(report['worst_tier'], 'r1z1-127.0.0.1')
        self.assertEqual(report['max_dispersion'], 6.593406593406593)
        # do a sencond rebalance
        rb.rebalance(seed=100)
        report = dispersion_report(rb, verbose=True)
        self.assertEqual(rb.dispersion, 16.666666666666668)
        self.assertEqual(report['worst_tier'], 'r1z0-127.0.0.3')
        self.assertEqual(report['max_dispersion'], 33.333333333333336)

        # ... but overload can square it
        rb.set_overload(rb.get_required_overload())
        rb.rebalance()
        self.assertEqual(rb.dispersion, 0.0)

    def test_parse_address_old_format(self):
        # Test old format
        argv = "127.0.0.1:6200R127.0.0.1:6200/sda1_some meta data"
        ip, port, rest = parse_address(argv)
        self.assertEqual(ip, '127.0.0.1')
        self.assertEqual(port, 6200)
        self.assertEqual(rest, 'R127.0.0.1:6200/sda1_some meta data')

    def test_normalized_device_tier_names(self):
        rb = ring.RingBuilder(8, 3, 0)
        rb.add_dev({
            'region': 1,
            'zone': 1,
            'ip': '127.0.0.1',
            'port': 6011,
            'device': 'd1',
            'weight': 0.0,
        })
        dev = rb.devs[0]
        expected = 'r1z1-127.0.0.1/d1'
        self.assertEqual(expected, get_tier_name(tiers_for_dev(dev)[-1], rb))
        self.assertEqual(expected, pretty_dev(dev))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Middleware that will provide Dynamic Large Object (DLO) support.

---------------
Using ``swift``
---------------

The quickest way to try out this feature is use the ``swift`` Swift Tool
included with the `python-swiftclient`_ library.  You can use the ``-S``
option to specify the segment size to use when splitting a large file. For
example::

    swift upload test_container -S 1073741824 large_file

This would split the large_file into 1G segments and begin uploading those
segments in parallel. Once all the segments have been uploaded, ``swift`` will
then create the manifest file so the segments can be downloaded as one.

So now, the following ``swift`` command would download the entire large
object::

    swift download test_container large_file

``swift`` command uses a strict convention for its segmented object
support. In the above example it will upload all the segments into a
second container named test_container_segments. These segments will
have names like large_file/1290206778.25/21474836480/00000000,
large_file/1290206778.25/21474836480/00000001, etc.

The main benefit for using a separate container is that the main container
listings will not be polluted with all the segment names. The reason for using
the segment name format of <name>/<timestamp>/<size>/<segment> is so that an
upload of a new file with the same name won't overwrite the contents of the
first until the last moment when the manifest file is updated.

``swift`` will manage these segment files for you, deleting old segments on
deletes and overwrites, etc. You can override this behavior with the
``--leave-segments`` option if desired; this is useful if you want to have
multiple versions of the same large object available.

.. _`python-swiftclient`: http://github.com/openstack/python-swiftclient

----------
Direct API
----------

You can also work with the segments and manifests directly with HTTP
requests instead of having ``swift`` do that for you. You can just
upload the segments like you would any other object and the manifest
is just a zero-byte (not enforced) file with an extra
``X-Object-Manifest`` header.

All the object segments need to be in the same container, have a common object
name prefix, and sort in the order in which they should be concatenated.
Object names are sorted lexicographically as UTF-8 byte strings.
They don't have to be in the same container as the manifest file will be, which
is useful to keep container listings clean as explained above with ``swift``.

The manifest file is simply a zero-byte (not enforced) file with the extra
``X-Object-Manifest: <container>/<prefix>`` header, where ``<container>`` is
the container the object segments are in and ``<prefix>`` is the common prefix
for all the segments.

It is best to upload all the segments first and then create or update the
manifest. In this way, the full object won't be available for downloading
until the upload is complete. Also, you can upload a new set of segments to
a second location and then update the manifest to point to this new location.
During the upload of the new segments, the original manifest will still be
available to download the first set of segments.

.. note::

    When updating a manifest object using a POST request, a
    ``X-Object-Manifest`` header must be included for the object to
    continue to behave as a manifest object.

    The manifest file should have no content. However, this is not enforced.
    If the manifest path itself conforms to container/prefix specified in
    ``X-Object-Manifest``, and if manifest has some content/data in it, it
    would also be considered as segment and manifest's content will be part of
    the concatenated GET response. The order of concatenation follows the usual
    DLO logic which is - the order of concatenation adheres to order returned
    when segment names are sorted.


Here's an example using ``curl`` with tiny 1-byte segments::

    # First, upload the segments
    curl -X PUT -H 'X-Auth-Token: <token>' \
        http://<storage_url>/container/myobject/00000001 --data-binary '1'
    curl -X PUT -H 'X-Auth-Token: <token>' \
        http://<storage_url>/container/myobject/00000002 --data-binary '2'
    curl -X PUT -H 'X-Auth-Token: <token>' \
        http://<storage_url>/container/myobject/00000003 --data-binary '3'

    # Next, create the manifest file
    curl -X PUT -H 'X-Auth-Token: <token>' \
        -H 'X-Object-Manifest: container/myobject/' \
        http://<storage_url>/container/myobject --data-binary ''

    # And now we can download the segments as a single object
    curl -H 'X-Auth-Token: <token>' \
        http://<storage_url>/container/myobject
"""

import json

from swift.common import constraints
from swift.common.exceptions import ListingIterError, SegmentError
from swift.common.http import is_success
from swift.common.swob import Request, Response, HTTPException, \
    HTTPRequestedRangeNotSatisfiable, HTTPBadRequest, HTTPConflict, \
    str_to_wsgi, wsgi_to_str, wsgi_quote, wsgi_unquote, normalize_etag
from swift.common.utils import get_logger, \
    RateLimitedIterator, quote, close_if_possible, closing_if_possible, \
    drain_and_close, md5
from swift.common.request_helpers import SegmentedIterable, \
    update_ignore_range_header
from swift.common.wsgi import WSGIContext, make_subrequest, load_app_config


class GetContext(WSGIContext):
    def __init__(self, dlo, logger):
        super(GetContext, self).__init__(dlo.app)
        self.dlo = dlo
        self.logger = logger

    def _get_container_listing(self, req, version, account, container,
                               prefix, marker=''):
        '''
        :param version: whatever
        :param account: native
        :param container: native
        :param prefix: native
        :param marker: native
        '''
        con_req = make_subrequest(
            req.environ,
            path=wsgi_quote('/'.join([
                '', str_to_wsgi(version),
                str_to_wsgi(account), str_to_wsgi(container)])),
            method='GET',
            headers={'x-auth-token': req.headers.get('x-auth-token')},
            agent=('%(orig)s ' + 'DLO MultipartGET'), swift_source='DLO')
        con_req.query_string = 'prefix=%s' % quote(prefix)
        if marker:
            con_req.query_string += '&marker=%s' % quote(marker)

        con_resp = con_req.get_response(self.dlo.app)
        if not is_success(con_resp.status_int):
            if req.method == 'HEAD':
                con_resp.body = b''
            return con_resp, None
        with closing_if_possible(con_resp.app_iter):
            return None, json.loads(b''.join(con_resp.app_iter))

    def _segment_listing_iterator(self, req, version, account, container,
                                  prefix, segments, first_byte=None,
                                  last_byte=None):
        '''
        :param req: upstream request
        :param version: native
        :param account: native
        :param container: native
        :param prefix: native
        :param segments: array of dicts, with native strings
        :param first_byte: number
        :param last_byte: number
        '''
        # It's sort of hokey that this thing takes in the first page of
        # segments as an argument, but we need to compute the etag and content
        # length from the first page, and it's better to have a hokey
        # interface than to make redundant requests.
        if first_byte is None:
            first_byte = 0
        if last_byte is None:
            last_byte = float("inf")

        while True:
            for segment in segments:
                seg_length = int(segment['bytes'])

                if first_byte >= seg_length:
                    # don't need any bytes from this segment
                    first_byte = max(first_byte - seg_length, -1)
                    last_byte = max(last_byte - seg_length, -1)
                    continue
                elif last_byte < 0:
                    # no bytes are needed from this or any future segment
                    break

                seg_name = segment['name']

                # We deliberately omit the etag and size here;
                # SegmentedIterable will check size and etag if
                # specified, but we don't want it to. DLOs only care
                # that the objects' names match the specified prefix.
                # SegmentedIterable will instead check that the data read
                # from each segment matches the response headers.
                _path = "/".join(["", version, account, container, seg_name])
                _first = None if first_byte <= 0 else first_byte
                _last = None if last_byte >= seg_length - 1 else last_byte
                yield {
                    'path': _path,
                    'first_byte': _first,
                    'last_byte': _last
                }

                first_byte = max(first_byte - seg_length, -1)
                last_byte = max(last_byte - seg_length, -1)

            if len(segments) < constraints.CONTAINER_LISTING_LIMIT:
                # a short page means that we're done with the listing
                break
            elif last_byte < 0:
                break

            marker = segments[-1]['name']
            error_response, segments = self._get_container_listing(
                req, version, account, container, prefix, marker)
            if error_response:
                # we've already started sending the response body to the
                # client, so all we can do is raise an exception to make the
                # WSGI server close the connection early
                close_if_possible(error_response.app_iter)
                raise ListingIterError(
                    "Got status %d listing container /%s/%s" %
                    (error_response.status_int, account, container))

    def get_or_head_response(self, req, x_object_manifest):
        '''
        :param req: user's request
        :param x_object_manifest: as unquoted, native string
        '''
        response_headers = self._response_headers

        container, obj_prefix = x_object_manifest.split('/', 1)

        version, account, _junk = req.split_path(2, 3, True)
        version = wsgi_to_str(version)
        account = wsgi_to_str(account)
        error_response, segments = self._get_container_listing(
            req, version, account, container, obj_prefix)
        if error_response:
            return error_response
        have_complete_listing = len(segments) < \
            constraints.CONTAINER_LISTING_LIMIT

        first_byte = last_byte = None
        actual_content_length = None
        content_length_for_swob_range = None
        if req.range and len(req.range.ranges) == 1:
            content_length_for_swob_range = sum(o['bytes'] for o in segments)

            # This is a hack to handle suffix byte ranges (e.g. "bytes=-5"),
            # which we can't honor unless we have a complete listing.
            _junk, range_end = req.range.ranges_for_length(float("inf"))[0]

            # If this is all the segments, we know whether or not this
            # range request is satisfiable.
            #
            # Alternately, we may not have all the segments, but this range
            # falls entirely within the first page's segments, so we know
            # that it is satisfiable.
            if (have_complete_listing
               or range_end < content_length_for_swob_range):
                byteranges = req.range.ranges_for_length(
                    content_length_for_swob_range)
                if not byteranges:
                    headers = {'Accept-Ranges': 'bytes'}
                    if have_complete_listing:
                        headers['Content-Range'] = 'bytes */%d' % (
                            content_length_for_swob_range, )
                    return HTTPRequestedRangeNotSatisfiable(
                        request=req, headers=headers)
                first_byte, last_byte = byteranges[0]
                # For some reason, swob.Range.ranges_for_length adds 1 to the
                # last byte's position.
                last_byte -= 1
                actual_content_length = last_byte - first_byte + 1
            else:
                # The range may or may not be satisfiable, but we can't tell
                # based on just one page of listing, and we're not going to go
                # get more pages because that would use up too many resources,
                # so we ignore the Range header and return the whole object.
                actual_content_length = None
                content_length_for_swob_range = None
                req.range = None
        else:
            req.range = None

        response_headers = [
            (h, v) for h, v in response_headers
            if h.lower() not in ("content-length", "content-range")]

        if content_length_for_swob_range is not None:
            # Here, we have to give swob a big-enough content length so that
            # it can compute the actual content length based on the Range
            # header. This value will not be visible to the client; swob will
            # substitute its own Content-Length.
            #
            # Note: if the manifest points to at least CONTAINER_LISTING_LIMIT
            # segments, this may be less than the sum of all the segments'
            # sizes. However, it'll still be greater than the last byte in the
            # Range header, so it's good enough for swob.
            response_headers.append(('Content-Length',
                                     str(content_length_for_swob_range)))
        elif have_complete_listing:
            actual_content_length = sum(o['bytes'] for o in segments)
            response_headers.append(('Content-Length',
                                     str(actual_content_length)))

        if have_complete_listing:
            response_headers = [(h, v) for h, v in response_headers
                                if h.lower() != "etag"]
            etag = md5(usedforsecurity=False)
            for seg_dict in segments:
                etag.update(normalize_etag(seg_dict['hash']).encode('utf8'))
            response_headers.append(('Etag', '"%s"' % etag.hexdigest()))

        app_iter = None
        if req.method == 'GET':
            listing_iter = RateLimitedIterator(
                self._segment_listing_iterator(
                    req, version, account, container, obj_prefix, segments,
                    first_byte=first_byte, last_byte=last_byte),
                self.dlo.rate_limit_segments_per_sec,
                limit_after=self.dlo.rate_limit_after_segment)

            app_iter = SegmentedIterable(
                req, self.dlo.app, listing_iter, ua_suffix="DLO MultipartGET",
                swift_source="DLO", name=req.path, logger=self.logger,
                max_get_time=self.dlo.max_get_time,
                response_body_length=actual_content_length)

            try:
                app_iter.validate_first_segment()
            except HTTPException as err_resp:
                return err_resp
            except (SegmentError, ListingIterError):
                return HTTPConflict(request=req)

        resp = Response(request=req, headers=response_headers,
                        conditional_response=True,
                        app_iter=app_iter)

        return resp

    def handle_request(self, req, start_response):
        """
        Take a GET or HEAD request, and if it is for a dynamic large object
        manifest, return an appropriate response.

        Otherwise, simply pass it through.
        """
        update_ignore_range_header(req, 'X-Object-Manifest')
        resp_iter = self._app_call(req.environ)

        # make sure this response is for a dynamic large object manifest
        for header, value in self._response_headers:
            if (header.lower() == 'x-object-manifest'):
                content_length = self._response_header_value('content-length')
                if content_length is not None and int(content_length) < 1024:
                    # Go ahead and consume small bodies
                    drain_and_close(resp_iter)
                close_if_possible(resp_iter)
                response = self.get_or_head_response(
                    req, wsgi_to_str(wsgi_unquote(value)))
                return response(req.environ, start_response)
        # Not a dynamic large object manifest; just pass it through.
        start_response(self._response_status,
                       self._response_headers,
                       self._response_exc_info)
        return resp_iter


class DynamicLargeObject(object):
    def __init__(self, app, conf):
        self.app = app
        self.logger = get_logger(conf, log_route='dlo')

        # DLO functionality used to live in the proxy server, not middleware,
        # so let's try to go find config values in the proxy's config section
        # to ease cluster upgrades.
        self._populate_config_from_old_location(conf)

        self.max_get_time = int(conf.get('max_get_time', '86400'))
        self.rate_limit_after_segment = int(conf.get(
            'rate_limit_after_segment', '10'))
        self.rate_limit_segments_per_sec = int(conf.get(
            'rate_limit_segments_per_sec', '1'))

    def _populate_config_from_old_location(self, conf):
        if ('rate_limit_after_segment' in conf or
                'rate_limit_segments_per_sec' in conf or
                'max_get_time' in conf or
                '__file__' not in conf):
            return

        proxy_conf = load_app_config(conf['__file__'])
        for setting in ('rate_limit_after_segment',
                        'rate_limit_segments_per_sec',
                        'max_get_time'):
            if setting in proxy_conf:
                conf[setting] = proxy_conf[setting]

    def __call__(self, env, start_response):
        """
        WSGI entry point
        """
        req = Request(env)
        try:
            vrs, account, container, obj = req.split_path(4, 4, True)
            is_obj_req = True
        except ValueError:
            is_obj_req = False
        if not is_obj_req:
            return self.app(env, start_response)

        if ((req.method == 'GET' or req.method == 'HEAD') and
                req.params.get('multipart-manifest') != 'get'):
            return GetContext(self, self.logger).\
                handle_request(req, start_response)
        elif req.method == 'PUT':
            error_response = self._validate_x_object_manifest_header(req)
            if error_response:
                return error_response(env, start_response)
        return self.app(env, start_response)

    def _validate_x_object_manifest_header(self, req):
        """
        Make sure that X-Object-Manifest is valid if present.
        """
        if 'X-Object-Manifest' in req.headers:
            value = req.headers['X-Object-Manifest']
            container = prefix = None
            try:
                container, prefix = value.split('/', 1)
            except ValueError:
                pass
            if not container or not prefix or '?' in value or '&' in value or \
                    prefix.startswith('/'):
                return HTTPBadRequest(
                    request=req,
                    body=('X-Object-Manifest must be in the '
                          'format container/prefix'))


def filter_factory(global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)

    def dlo_filter(app):
        return DynamicLargeObject(app, conf)
    return dlo_filter



# Copyright (c) 2010-2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from swift.common.http import HTTP_OK
from swift.common.middleware.acl import parse_acl, referrer_allowed
from swift.common.utils import public

from swift.common.middleware.s3api.exception import ACLError
from swift.common.middleware.s3api.controllers.base import Controller
from swift.common.middleware.s3api.s3response import (
    HTTPOk, S3NotImplemented, MalformedACLError, UnexpectedContent,
    MissingSecurityHeader)
from swift.common.middleware.s3api.etree import Element, SubElement, tostring
from swift.common.middleware.s3api.acl_utils import swift_acl_translate, \
    XMLNS_XSI


MAX_ACL_BODY_SIZE = 200 * 1024


def get_acl(account_name, headers):
    """
    Attempts to construct an S3 ACL based on what is found in the swift headers
    """

    elem = Element('AccessControlPolicy')
    owner = SubElement(elem, 'Owner')
    SubElement(owner, 'ID').text = account_name
    SubElement(owner, 'DisplayName').text = account_name
    access_control_list = SubElement(elem, 'AccessControlList')

    # grant FULL_CONTROL to myself by default
    grant = SubElement(access_control_list, 'Grant')
    grantee = SubElement(grant, 'Grantee', nsmap={'xsi': XMLNS_XSI})
    grantee.set('{%s}type' % XMLNS_XSI, 'CanonicalUser')
    SubElement(grantee, 'ID').text = account_name
    SubElement(grantee, 'DisplayName').text = account_name
    SubElement(grant, 'Permission').text = 'FULL_CONTROL'

    referrers, _ = parse_acl(headers.get('x-container-read'))
    if referrer_allowed('unknown', referrers):
        # grant public-read access
        grant = SubElement(access_control_list, 'Grant')
        grantee = SubElement(grant, 'Grantee', nsmap={'xsi': XMLNS_XSI})
        grantee.set('{%s}type' % XMLNS_XSI, 'Group')
        SubElement(grantee, 'URI').text = \
            'http://acs.amazonaws.com/groups/global/AllUsers'
        SubElement(grant, 'Permission').text = 'READ'

    referrers, _ = parse_acl(headers.get('x-container-write'))
    if referrer_allowed('unknown', referrers):
        # grant public-write access
        grant = SubElement(access_control_list, 'Grant')
        grantee = SubElement(grant, 'Grantee', nsmap={'xsi': XMLNS_XSI})
        grantee.set('{%s}type' % XMLNS_XSI, 'Group')
        SubElement(grantee, 'URI').text = \
            'http://acs.amazonaws.com/groups/global/AllUsers'
        SubElement(grant, 'Permission').text = 'WRITE'

    body = tostring(elem)

    return HTTPOk(body=body, content_type="text/plain")


class AclController(Controller):
    """
    Handles the following APIs:

    * GET Bucket acl
    * PUT Bucket acl
    * GET Object acl
    * PUT Object acl

    Those APIs are logged as ACL operations in the S3 server log.
    """
    @public
    def GET(self, req):
        """
        Handles GET Bucket acl and GET Object acl.
        """
        resp = req.get_response(self.app, method='HEAD')

        return get_acl(req.user_id, resp.headers)

    @public
    def PUT(self, req):
        """
        Handles PUT Bucket acl and PUT Object acl.
        """
        if req.is_object_request:
            # Handle Object ACL
            raise S3NotImplemented()
        else:
            # Handle Bucket ACL
            xml = req.xml(MAX_ACL_BODY_SIZE)
            if all(['HTTP_X_AMZ_ACL' in req.environ, xml]):
                # S3 doesn't allow to give ACL with both ACL header and body.
                raise UnexpectedContent()
            elif not any(['HTTP_X_AMZ_ACL' in req.environ, xml]):
                # Both canned ACL header and xml body are missing
                raise MissingSecurityHeader(missing_header_name='x-amz-acl')
            else:
                # correct ACL exists in the request
                if xml:
                    # We very likely have an XML-based ACL request.
                    # let's try to translate to the request header
                    try:
                        translated_acl = swift_acl_translate(xml, xml=True)
                    except ACLError:
                        raise MalformedACLError()

                    for header, acl in translated_acl:
                        req.headers[header] = acl

            resp = req.get_response(self.app, 'POST')
            resp.status = HTTP_OK
            resp.headers.update({'Location': req.container_name})

            return resp



# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import dataclasses
import io
import json
import os.path
import unittest
from unittest import mock

from swift.common.ring.io import IndexEntry, RingReader, RingWriter

from test.unit import with_tempdir


class TestRoundTrip(unittest.TestCase):
    def assertRepeats(self, data, pattern, n):
        l = len(pattern)
        self.assertEqual(len(data), n * l)
        actual = collections.Counter(
            data[x * l:(x + 1) * l]
            for x in range(n))
        self.assertEqual(actual, {pattern: n})

    @with_tempdir
    def test_write_failure(self, tempd):
        tempf = os.path.join(tempd, 'not-persisted')
        try:
            with RingWriter.open(tempf):
                self.assertEqual(1, len(os.listdir(tempd)))
                raise RuntimeError
        except RuntimeError:
            pass
        self.assertEqual(0, len(os.listdir(tempd)))

    def test_sections(self):
        buf = io.BytesIO()
        with RingWriter(buf) as writer:
            writer.write_magic(2)
            with writer.section('foo'):
                writer.write_blob(b'\xde\xad\xbe\xef' * 10240)

            with writer.section('bar'):
                # Sometimes you might not want to get the whole section into
                # memory as a byte-string all at once (eg, when writing ring
                # assignments)
                writer.write_size(40960)
                for _ in range(10):
                    writer.write(b'\xda\x7a\xda\x7a' * 1024)

            with writer.section('baz'):
                writer.write_blob(b'more' * 10240)

                # Can't nest sections
                with self.assertRaises(ValueError):
                    with writer.section('inner'):
                        pass
                self.assertNotIn('inner', writer.index)

            writer.write(b'can add arbitrary bytes')
            # ...though accessing them on read may be difficult; see below.
            # This *is not* a recommended pattern -- write proper length-value
            # blobs instead (even if you don't include them as sections in the
            # index).

            with writer.section('quux'):
                writer.write_blob(b'data' * 10240)

            # Gotta do this at the start
            with self.assertRaises(IOError):
                writer.write_magic(2)

            # Can't write duplicate sections
            with self.assertRaises(ValueError):
                with writer.section('foo'):
                    pass

            # We're reserving globs, so we can later support something like
            # reader.load_sections('swift/ring/*')
            with self.assertRaises(ValueError):
                with writer.section('foo*'):
                    pass

        buf.seek(0)
        reader = RingReader(buf)
        self.assertEqual(reader.version, 2)
        # Order matters!
        self.assertEqual(list(reader.index), [
            'foo', 'bar', 'baz', 'quux'])
        self.assertEqual({
            k: (v.uncompressed_start, v.uncompressed_end, v.checksum_method)
            for k, v in reader.index.items()
        }, {
            'foo': (6, 40974, 'sha256'),
            'bar': (40974, 81942, 'sha256'),
            'baz': (81942, 122910, 'sha256'),
            # note the gap between baz and quux for the raw bytes
            'quux': (122933, 163901, 'sha256'),
        })

        self.assertIn('foo', reader)
        self.assertNotIn('inner', reader)

        self.assertRepeats(reader.read_section('foo'),
                           b'\xde\xad\xbe\xef', 10240)
        with reader.open_section('bar') as s:
            for _ in range(10):
                self.assertEqual(s.read(4), b'\xda\x7a\xda\x7a')
            self.assertRepeats(s.read(), b'\xda\x7a\xda\x7a', 10230)
        # If you know that one section follows another, you don't *have*
        # to "open" the next one
        self.assertRepeats(reader.read_blob(), b'more', 10240)
        self.assertRepeats(reader.read_section('quux'),
                           b'data', 10240)
        # Index is just a final (length-prefixed) JSON blob
        index_dict = json.loads(reader.read_blob())
        self.assertEqual(reader.index, {
            section: IndexEntry(*entry)
            for section, entry in index_dict.items()})

        # Missing section
        with self.assertRaises(KeyError) as caught:
            with reader.open_section('foobar'):
                pass
        self.assertEqual("'foobar'", str(caught.exception))

        # seek to the end of baz
        reader.compressed_seek(reader.index['baz'].compressed_end)
        # so we can read the raw bytes we stuffed in
        gap_length = (reader.index['quux'].uncompressed_start -
                      reader.index['baz'].uncompressed_end)
        self.assertGreater(gap_length, 0)
        self.assertEqual(b'can add arbitrary bytes',
                         reader.read(gap_length))

    def test_sections_with_corruption(self):
        buf = io.BytesIO()
        with RingWriter(buf) as writer:
            writer.write_magic(2)
            with writer.section('foo'):
                writer.write_blob(b'\xde\xad\xbe\xef' * 10240)

        buf.seek(0)
        reader = RingReader(buf)
        # if you open a section, you better read it all!
        read_bytes = b''
        with self.assertRaises(ValueError) as caught:
            with reader.open_section('foo') as s:
                read_bytes = s.read(4)
        self.assertEqual(
            'Incomplete read; expected 40956 more bytes to be read',
            str(caught.exception))
        self.assertEqual(b'\xde\xad\xbe\xef', read_bytes)

        # if there's a digest mismatch, you can read data, but it'll
        # throw an error on close
        self.assertEqual('sha256', reader.index['foo'].checksum_method)
        self.assertEqual(
            'c51d6703d54cd7cf57b4d4b7ecfcca60'
            '56dbd41ebf1c1e83c0e8e48baeff629a',
            reader.index['foo'].checksum_value)
        reader.index['foo'] = dataclasses.replace(
            writer.index['foo'],
            checksum_value='not-the-sha',
        )
        read_bytes = b''
        with self.assertRaises(ValueError) as caught:
            with reader.open_section('foo') as s:
                read_bytes = s.read()
        self.assertIn('Hash mismatch in block: ', str(caught.exception))
        self.assertRepeats(read_bytes, b'\xde\xad\xbe\xef', 10240)

    @mock.patch('logging.getLogger')
    def test_sections_with_unsupported_checksum(self, mock_logging):
        buf = io.BytesIO()
        with RingWriter(buf) as writer:
            writer.write_magic(2)
            with writer.section('foo'):
                writer.write_blob(b'\xde\xad\xbe\xef')
            writer.index['foo'] = dataclasses.replace(
                writer.index['foo'],
                checksum_method='not_a_digest',
                checksum_value='do not care',
            )

        buf.seek(0)
        reader = RingReader(buf)
        with self.assertRaises(ValueError):
            with reader.open_section('foo'):
                pass

    def test_recompressed(self):
        buf = io.BytesIO()
        with RingWriter(buf) as writer:
            writer.write_magic(2)
            with writer.section('foo'):
                writer.write_blob(b'\xde\xad\xbe\xef' * 10240)

        buf.seek(0)
        reader = RingReader(buf)
        with self.assertRaises(IOError):
            reader.read(-1)  # don't be greedy
        uncompressed_bytes = reader.read(2 ** 20)

        buf = io.BytesIO()
        with RingWriter(buf) as writer:
            writer.write(uncompressed_bytes)

        buf.seek(0)
        with self.assertRaises(IOError):
            # ...but we can't read it
            RingReader(buf)

    def test_version_too_high(self):
        buf = io.BytesIO()
        with RingWriter(buf) as writer:
            # you can write it...
            writer.write_magic(3)
            with writer.section('foo'):
                writer.write_blob(b'\xde\xad\xbe\xef' * 10240)

        buf.seek(0)
        with self.assertRaises(ValueError):
            # ...but we can't read it
            RingReader(buf)



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Tests for `swift.common.splice`'''

import os
import errno
import ctypes
import logging
import tempfile
import unittest
import contextlib
import re

from unittest import mock

from swift.common.splice import splice, tee

LOGGER = logging.getLogger(__name__)


def safe_close(fd):
    '''Close a file descriptor, ignoring any exceptions'''

    try:
        os.close(fd)
    except Exception:
        LOGGER.exception('Error while closing FD')


@contextlib.contextmanager
def pipe():
    '''Context-manager providing 2 ends of a pipe, closing them at exit'''

    fds = os.pipe()

    try:
        yield fds
    finally:
        safe_close(fds[0])
        safe_close(fds[1])


class TestSplice(unittest.TestCase):
    '''Tests for `splice`'''

    def setUp(self):
        if not splice.available:
            raise unittest.SkipTest('splice not available')

    def test_flags(self):
        '''Test flag attribute availability'''

        self.assertTrue(hasattr(splice, 'SPLICE_F_MOVE'))
        self.assertTrue(hasattr(splice, 'SPLICE_F_NONBLOCK'))
        self.assertTrue(hasattr(splice, 'SPLICE_F_MORE'))
        self.assertTrue(hasattr(splice, 'SPLICE_F_GIFT'))

    @mock.patch('swift.common.splice.splice._c_splice', None)
    def test_available(self):
        '''Test `available` attribute correctness'''

        self.assertFalse(splice.available)

    def test_splice_pipe_to_pipe(self):
        '''Test `splice` from a pipe to a pipe'''

        with pipe() as (p1a, p1b):
            with pipe() as (p2a, p2b):
                os.write(p1b, b'abcdef')
                res = splice(p1a, None, p2b, None, 3, 0)
                self.assertEqual(res, (3, None, None))
                self.assertEqual(os.read(p2a, 3), b'abc')
                self.assertEqual(os.read(p1a, 3), b'def')

    def test_splice_file_to_pipe(self):
        '''Test `splice` from a file to a pipe'''

        with tempfile.NamedTemporaryFile(buffering=0) as fd:
            with pipe() as (pa, pb):
                fd.write(b'abcdef')
                fd.seek(0, os.SEEK_SET)

                res = splice(fd, None, pb, None, 3, 0)
                self.assertEqual(res, (3, None, None))
                # `fd.tell()` isn't updated...
                self.assertEqual(os.lseek(fd.fileno(), 0, os.SEEK_CUR), 3)

                fd.seek(0, os.SEEK_SET)
                res = splice(fd, 3, pb, None, 3, 0)
                self.assertEqual(res, (3, 6, None))
                self.assertEqual(os.lseek(fd.fileno(), 0, os.SEEK_CUR), 0)

                self.assertEqual(os.read(pa, 6), b'abcdef')

    def test_splice_pipe_to_file(self):
        '''Test `splice` from a pipe to a file'''

        with tempfile.NamedTemporaryFile(buffering=0) as fd:
            with pipe() as (pa, pb):
                os.write(pb, b'abcdef')

                res = splice(pa, None, fd, None, 3, 0)
                self.assertEqual(res, (3, None, None))
                self.assertEqual(fd.tell(), 3)

                fd.seek(0, os.SEEK_SET)

                res = splice(pa, None, fd, 3, 3, 0)
                self.assertEqual(res, (3, None, 6))
                self.assertEqual(fd.tell(), 0)

                self.assertEqual(fd.read(6), b'abcdef')

    @mock.patch.object(splice, '_c_splice')
    def test_fileno(self, mock_splice):
        '''Test handling of file-descriptors'''

        splice(1, None, 2, None, 3, 0)
        self.assertEqual(mock_splice.call_args,
                         ((1, None, 2, None, 3, 0), {}))

        mock_splice.reset_mock()

        with open('/dev/zero', 'r') as fd:
            splice(fd, None, fd, None, 3, 0)
            self.assertEqual(mock_splice.call_args,
                             ((fd.fileno(), None, fd.fileno(), None, 3, 0),
                              {}))

    @mock.patch.object(splice, '_c_splice')
    def test_flags_list(self, mock_splice):
        '''Test handling of flag lists'''

        splice(1, None, 2, None, 3,
               [splice.SPLICE_F_MOVE, splice.SPLICE_F_NONBLOCK])

        flags = splice.SPLICE_F_MOVE | splice.SPLICE_F_NONBLOCK
        self.assertEqual(mock_splice.call_args,
                         ((1, None, 2, None, 3, flags), {}))

        mock_splice.reset_mock()

        splice(1, None, 2, None, 3, [])
        self.assertEqual(mock_splice.call_args,
                         ((1, None, 2, None, 3, 0), {}))

    def test_errno(self):
        '''Test handling of failures'''

        # Invoke EBADF by using a read-only FD as fd_out
        with open('/dev/null', 'r') as fd:
            err = errno.EBADF
            msg = r'\[Errno %d\] splice: %s' % (err, os.strerror(err))
            try:
                splice(fd, None, fd, None, 3, 0)
            except IOError as e:
                self.assertTrue(re.match(msg, str(e)))
            else:
                self.fail('Expected IOError was not raised')

        self.assertEqual(ctypes.get_errno(), 0)

    @mock.patch('swift.common.splice.splice._c_splice', None)
    def test_unavailable(self):
        '''Test exception when unavailable'''

        self.assertRaises(EnvironmentError, splice, 1, None, 2, None, 2, 0)

    def test_unavailable_in_libc(self):
        '''Test `available` attribute when `libc` has no `splice` support'''

        class LibC(object):
            '''A fake `libc` object tracking `splice` attribute access'''

            def __init__(self):
                self.splice_retrieved = False

            @property
            def splice(self):
                self.splice_retrieved = True
                raise AttributeError

        libc = LibC()
        mock_cdll = mock.Mock(return_value=libc)

        with mock.patch('ctypes.CDLL', new=mock_cdll):
            # Force re-construction of a `Splice` instance
            # Something you're not supposed to do in actual code
            new_splice = type(splice)()
            self.assertFalse(new_splice.available)

        libc_name = ctypes.util.find_library('c')

        mock_cdll.assert_called_once_with(libc_name, use_errno=True)
        self.assertTrue(libc.splice_retrieved)


class TestTee(unittest.TestCase):
    '''Tests for `tee`'''

    def setUp(self):
        if not tee.available:
            raise unittest.SkipTest('tee not available')

    @mock.patch('swift.common.splice.tee._c_tee', None)
    def test_available(self):
        '''Test `available` attribute correctness'''

        self.assertFalse(tee.available)

    def test_tee_pipe_to_pipe(self):
        '''Test `tee` from a pipe to a pipe'''

        with pipe() as (p1a, p1b):
            with pipe() as (p2a, p2b):
                os.write(p1b, b'abcdef')
                res = tee(p1a, p2b, 3, 0)
                self.assertEqual(res, 3)
                self.assertEqual(os.read(p2a, 3), b'abc')
                self.assertEqual(os.read(p1a, 6), b'abcdef')

    @mock.patch.object(tee, '_c_tee')
    def test_fileno(self, mock_tee):
        '''Test handling of file-descriptors'''

        with pipe() as (pa, pb):
            tee(pa, pb, 3, 0)
            self.assertEqual(mock_tee.call_args, ((pa, pb, 3, 0), {}))

            mock_tee.reset_mock()

            tee(os.fdopen(pa, 'r'), os.fdopen(pb, 'w'), 3, 0)
            self.assertEqual(mock_tee.call_args, ((pa, pb, 3, 0), {}))

    @mock.patch.object(tee, '_c_tee')
    def test_flags_list(self, mock_tee):
        '''Test handling of flag lists'''

        tee(1, 2, 3, [splice.SPLICE_F_MOVE | splice.SPLICE_F_NONBLOCK])
        flags = splice.SPLICE_F_MOVE | splice.SPLICE_F_NONBLOCK
        self.assertEqual(mock_tee.call_args, ((1, 2, 3, flags), {}))

        mock_tee.reset_mock()

        tee(1, 2, 3, [])
        self.assertEqual(mock_tee.call_args, ((1, 2, 3, 0), {}))

    def test_errno(self):
        '''Test handling of failures'''

        # Invoke EBADF by using a read-only FD as fd_out
        with open('/dev/null', 'r') as fd:
            err = errno.EBADF
            msg = r'\[Errno %d\] tee: %s' % (err, os.strerror(err))
            try:
                tee(fd, fd, 3, 0)
            except IOError as e:
                self.assertTrue(re.match(msg, str(e)))
            else:
                self.fail('Expected IOError was not raised')

        self.assertEqual(ctypes.get_errno(), 0)

    @mock.patch('swift.common.splice.tee._c_tee', None)
    def test_unavailable(self):
        '''Test exception when unavailable'''

        self.assertRaises(EnvironmentError, tee, 1, 2, 2, 0)

    def test_unavailable_in_libc(self):
        '''Test `available` attribute when `libc` has no `tee` support'''

        class LibC(object):
            '''A fake `libc` object tracking `tee` attribute access'''

            def __init__(self):
                self.tee_retrieved = False

            @property
            def tee(self):
                self.tee_retrieved = True
                raise AttributeError

        libc = LibC()
        mock_cdll = mock.Mock(return_value=libc)

        with mock.patch('ctypes.CDLL', new=mock_cdll):
            # Force re-construction of a `Tee` instance
            # Something you're not supposed to do in actual code
            new_tee = type(tee)()
            self.assertFalse(new_tee.available)

        libc_name = ctypes.util.find_library('c')

        mock_cdll.assert_called_once_with(libc_name, use_errno=True)
        self.assertTrue(libc.tee_retrieved)



# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import binascii
import hmac

from swift.common.utils import strict_b64decode


DEFAULT_ALLOWED_DIGESTS = 'sha1 sha256 sha512'
DEPRECATED_DIGESTS = {'sha1'}
SUPPORTED_DIGESTS = set(DEFAULT_ALLOWED_DIGESTS.split()) | DEPRECATED_DIGESTS


def get_hmac(request_method, path, expires, key, digest="sha1",
             ip_range=None):
    """
    Returns the hexdigest string of the HMAC (see RFC 2104) for
    the request.

    :param request_method: Request method to allow.
    :param path: The path to the resource to allow access to.
    :param expires: Unix timestamp as an int for when the URL
                    expires.
    :param key: HMAC shared secret.
    :param digest: constructor or the string name for the digest to use in
                   calculating the HMAC
                   Defaults to SHA1
    :param ip_range: The ip range from which the resource is allowed
                     to be accessed. We need to put the ip_range as the
                     first argument to hmac to avoid manipulation of the path
                     due to newlines being valid in paths
                     e.g. /v1/a/c/o\\n127.0.0.1
    :returns: hexdigest str of the HMAC for the request using the specified
              digest algorithm.
    """
    # These are the three mandatory fields.
    parts = [request_method, str(expires), path]
    formats = [b"%s", b"%s", b"%s"]

    if ip_range:
        parts.insert(0, ip_range)
        formats.insert(0, b"ip=%s")

    if isinstance(key, str):
        key = key.encode('utf8')

    message = b'\n'.join(
        fmt % (part if isinstance(part, bytes)
               else part.encode("utf-8"))
        for fmt, part in zip(formats, parts))

    return hmac.new(key, message, digest).hexdigest()


def get_allowed_digests(conf_digests, logger=None):
    """
    Pulls out 'allowed_digests' from the supplied conf. Then compares them with
    the list of supported and deprecated digests and returns whatever remain.

    When something is unsupported or deprecated it'll log a warning.

    :param conf_digests: iterable of allowed digests. If empty, defaults to
        DEFAULT_ALLOWED_DIGESTS.
    :param logger: optional logger; if provided, use it issue deprecation
        warnings
    :returns: A set of allowed digests that are supported and a set of
        deprecated digests.
    :raises: ValueError, if there are no digests left to return.
    """
    allowed_digests = set(digest.lower() for digest in conf_digests)
    if not allowed_digests:
        allowed_digests = SUPPORTED_DIGESTS

    not_supported = allowed_digests - SUPPORTED_DIGESTS
    if not_supported:
        if logger:
            logger.warning('The following digest algorithms are configured '
                           'but not supported: %s', ', '.join(not_supported))
        allowed_digests -= not_supported
    deprecated = allowed_digests & DEPRECATED_DIGESTS
    if deprecated and logger:
        if not conf_digests:
            logger.warning('The following digest algorithms are allowed by '
                           'default but deprecated: %s. Support will be '
                           'disabled by default in a future release, and '
                           'later removed entirely.', ', '.join(deprecated))
        else:
            logger.warning('The following digest algorithms are configured '
                           'but deprecated: %s. Support will be removed in a '
                           'future release.', ', '.join(deprecated))
    if not allowed_digests:
        raise ValueError('No valid digest algorithms are configured')

    return allowed_digests, deprecated


def extract_digest_and_algorithm(value):
    """
    Returns a tuple of (digest_algorithm, hex_encoded_digest)
    from a client-provided string of the form::

       <hex-encoded digest>

    or::

       <algorithm>:<base64-encoded digest>

    Note that hex-encoded strings must use one of sha1, sha256, or sha512.

    :raises: ValueError on parse failures
    """
    if ':' in value:
        algo, value = value.split(':', 1)
        # accept both standard and url-safe base64
        if ('-' in value or '_' in value) and not (
                '+' in value or '/' in value):
            value = value.replace('-', '+').replace('_', '/')
        value = binascii.hexlify(
            strict_b64decode(value + '==')).decode('ascii')
    else:
        binascii.unhexlify(value)  # make sure it decodes
        algo = {
            40: 'sha1',
            64: 'sha256',
            128: 'sha512',
        }.get(len(value))
        if not algo:
            raise ValueError('Bad digest length')
    return algo, value
